/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.client.gui;

import java.awt.Component;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.JList;


/**
 * @author Ian Thompson
 * 
 */
public class ConnectionsList extends JList {

	private final DefaultListModel model;

	private final File file;

	/**
	 * @param file
	 * 
	 */
	public ConnectionsList(File file) {
		super();
		this.model = new DefaultListModel();
		this.file = file;
		this.setModel(model);
		this.setCellRenderer(new CellRenderer());
	}

	/**
	 * Loads file of the format
	 * 
	 * <pre>
	 * [Connection Name]
	 * host=Hostname or IP Address
	 * port=Port Number
	 * [Second Connection]
	 * ...
	 * </pre>
	 * 
	 * Also accepts blank lines and lines starting with the <code>'#'</code>
	 * character.
	 * 
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws IllegalArgumentException
	 */
	public void load() throws FileNotFoundException, IOException,
			IllegalArgumentException {
		BufferedReader fileIn = new BufferedReader(new FileReader(file));
		String line;
		ConnectionItem item = null;
		model.clear();
		while ((line = fileIn.readLine()) != null) {
			if (line.startsWith("#") || line.equals("")) {
				continue;
			} else if (line.startsWith("[") && line.endsWith("]")) {
				if (item != null)
					model.addElement(item);
				item = new ConnectionItem();
				item.name = line.substring(1, line.length() - 1);
			} else if (line.startsWith("host=")) {
				item.host = line.substring(5);
			} else if (line.startsWith("port=")) {
				item.port = Integer.parseInt(line.substring(5));
			} else {
				throw new IllegalArgumentException("Unsupported line type.");
			}
		}
		if (item != null)
			model.addElement(item);

		fileIn.close();
	}

	/**
	 * 
	 * @throws IOException
	 */
	public void save() throws IOException {
		PrintWriter fileOut = new PrintWriter(new FileWriter(file));
		fileOut.println("# Stored Connections List");
		fileOut.println();
		ConnectionItem item;
		for (Object o : model.toArray()) {
			if (o instanceof ConnectionItem) {
				item = (ConnectionItem) o;
				fileOut.println("[" + item.name + "]");
				fileOut.println("host=" + item.host);
				fileOut.println("port=" + item.port);
				fileOut.println();
			}
		}
		fileOut.close();
	}

	/**
	 * 
	 * @param name
	 * @param host
	 * @param port
	 */
	public void addConnection(String name, String host, int port) {
		ConnectionItem ci = new ConnectionItem();
		ci.name = name;
		ci.host = host;
		ci.port = port;
		model.addElement(ci);
	}

	/**
	 * Removes the selected entry from the list.
	 */
	public void removeSelected() {
		Object obj = getSelectedValue();
		if (obj != null)
			model.removeElement(obj);
	}

	/**
	 * 
	 * @return String
	 */
	public String getSelectedName() {
		Object obj = getSelectedValue();
		if (obj instanceof ConnectionItem)
			return ((ConnectionItem) obj).name;
		return null;
	}

	/**
	 * @param name
	 */
	public void setSelectedName(String name) {
		Object obj = getSelectedValue();
		if (obj instanceof ConnectionItem)
			((ConnectionItem) obj).name = name;
		repaint();
	}

	/**
	 * 
	 * @return String
	 */
	public String getSelectedAddress() {
		Object obj = getSelectedValue();
		if (obj instanceof ConnectionItem)
			return ((ConnectionItem) obj).host;
		return null;
	}

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class CellRenderer extends DefaultListCellRenderer {

		@Override
		public Component getListCellRendererComponent(JList list, Object value,
				int index, boolean isSelected, boolean cellHasFocus) {

			Component c = super.getListCellRendererComponent(list, value,
					index, isSelected, cellHasFocus);
			if (value instanceof ConnectionItem) {
				ConnectionItem item = (ConnectionItem) value;
				setText(item.toString());
				setIcon((item.connected ? Icons.CONNECTION_CONNECTED
						: Icons.CONNECTION_DISCONNECTED).getImageIcon());
				return this;
			}
			return c;
		}
	}

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class ConnectionItem {

		String name;

		String host;

		int port;

		boolean connected;

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return name + " <" + host + ">";
		}
	}
}
