/*
 * Distriblend: A distributed rendering system for Blender
 * Copyright (C) 2007: Ian Philip Thompson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.client.gui;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.TreeSet;

import javax.swing.AbstractListModel;

/**
 * @author Ian Thompson
 * 
 */
public class RangesListModel extends AbstractListModel {

	/**
	 * 
	 */
	private LinkedList<Range> ranges;

	/**
	 * 
	 */
	public RangesListModel() {
		ranges = new LinkedList<Range>();
	}

	/**
	 * Adds a range starting at <code>lower</code> stepping by
	 * <code>step</code> until <code>upper</code> is reached.
	 * 
	 * 
	 * @param lower
	 *            The lower limit of the range to add.
	 * @param upper
	 *            The upper limit of the range to add.
	 * @param step
	 *            The interval between each value in the range.
	 */
	public void addRange(int lower, int upper, int step) {

		Range r = new Range(lower, upper, step);
		int i = 0;
		for (Range s : ranges) {
			if (r.compareTo(s) > 0)
				i++;
			else {
				System.out.println(i);
				ranges.add(i, r);
				fireIntervalAdded(this, i, i);
				return;
			}
		}
		ranges.add(r);
		fireIntervalAdded(this, i, i);
	}

	/**
	 * Removes the element from the list at the specified index.
	 * 
	 * @param index
	 *            The index of the element to remove.
	 */
	public void remove(int index) {
		ranges.remove(index);
		fireIntervalRemoved(this, index, index);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.ListModel#getElementAt(int)
	 */
	public Object getElementAt(int index) {
		Range r = ranges.get(index);
		int lower = r.lower;
		int upper = r.upper;
		int step = r.step;

		assert step > 0;
		assert lower <= upper;

		if (lower == upper)
			return "" + lower;
		else if (step == 1)
			return lower + " to " + upper;
		else
			return lower + " to " + upper + " step " + step;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.ListModel#getSize()
	 */
	public int getSize() {
		return ranges.size();
	}
	
	/**
	 * @return Ranges
	 */
	public Iterator<Range> ranges() {
		return ranges.iterator();
	}

	/**
	 * 
	 * @return Frames
	 */
	public int[] getFrames() {
		TreeSet<Integer> frames = new TreeSet<Integer>();
		for (Range r : ranges) {
			for (int i = r.lower; i <= r.upper; i += r.step)
				frames.add(i);
		}
		int[] ints = new int[frames.size()];
		int i = 0;
		for (Integer frame : frames)
			ints[i++] = frame.intValue();
		return ints;
	}

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public class Range implements Comparable<Range> {
		int lower, upper, step;

		/**
		 * 
		 * @param lower
		 * @param upper
		 * @param step
		 */
		public Range(int lower, int upper, int step) {
			this.lower = lower;
			this.upper = upper;
			this.step = step;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(Range o) {
			int diff = this.lower - o.lower;
			if (diff == 0)
				diff = this.upper - o.upper;
			if (diff == 0)
				diff = this.step - o.step;
			return diff;
		}
	}
}
