/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.distributor;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.distributor.storage.Project;
import net.sf.distriblend.distributor.storage.Storage;
import net.sf.distriblend.distributor.storage.StorageException;
import net.sf.distriblend.distributor.storage.User;
import net.sf.distriblend.distributor.storage.Project.Archive;
import net.sf.distriblend.share.FileMap;
import net.sf.distriblend.share.LocalPath;
import net.sf.distriblend.share.Settings;
import net.sf.distriblend.share.Task;
import net.sf.distriblend.share.net.ConnectionHandler;
import net.sf.distriblend.share.net.Receptor;
import net.sf.distriblend.share.protocol.Authorization;
import net.sf.distriblend.share.protocol.FileTransfer;
import net.sf.distriblend.share.protocol.Submission.CreateProjectRequest;
import net.sf.distriblend.share.protocol.Submission.CreateProjectResponse;
import net.sf.distriblend.share.protocol.Submission.DeleteRequest;
import net.sf.distriblend.share.protocol.Submission.DeleteResponse;
import net.sf.distriblend.share.protocol.Submission.FramesRequest;
import net.sf.distriblend.share.protocol.Submission.FramesResponse;
import net.sf.distriblend.share.protocol.Submission.ListProjectsRequest;
import net.sf.distriblend.share.protocol.Submission.ListProjectsResponse;
import net.sf.distriblend.share.protocol.Submission.MonitorItem;
import net.sf.distriblend.share.protocol.Submission.MonitorTasksRequest;
import net.sf.distriblend.share.protocol.Submission.MonitorTasksResponse;
import net.sf.distriblend.share.protocol.Submission.SubmitFilesRequest;
import net.sf.distriblend.share.protocol.Submission.SubmitFilesResponse;
import net.sf.distriblend.share.protocol.Submission.SubmitTasksRequest;
import net.sf.distriblend.share.protocol.Submission.SubmitTasksResponse;

/**
 * @author Ian Thompson
 * 
 */
public class ClientHandler extends ConnectionHandler {

	/**
	 * 
	 */
	public ClientHandler() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see net.sf.distriblend.share.net.ConnectionHandler#handleConnection(java.net.Socket)
	 */
	@Override
	public void handleConnection(Socket socket) throws IOException {
		logger.log(Level.INFO, "Client connected from {0}.", socket
				.getInetAddress().getHostName());
		if (thread == null) {
			int timeout = Settings.instance().getIntProperty(
					"server.client.timeout");
			this.receptor = new Receptor(socket, timeout);
			this.thread = new ClientHandlerThread();
			this.thread.start();
		}
	}

	/**
	 * Forcibly disconnects the client.
	 */
	public void disconnect() {
		receptor.close();
	}

	/* -- Private -- */

	/**
	 * 
	 */
	private static Logger logger = Logger.getLogger(ClientHandler.class
			.getName());

	/**
	 * 
	 */
	private ClientHandlerThread thread = null;

	/**
	 * 
	 */
	private Receptor receptor;

	private Project currentProject = null;

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class ClientHandlerThread extends Thread {

		/**
		 * 
		 * @throws IOException
		 */
		public ClientHandlerThread() throws IOException {
			super("ClientHandlerWorker");
			this.setDaemon(true);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			while (true) {
				try {
					Object o = receptor.receive();

					/* Authorize request. */
					if (o instanceof Authorization.Request) {
						Authorization.Request request = (Authorization.Request) o;
						String userName = request.getUser();

						/* Send salt for encoding password. */
						String salt = Authorization.generateSalt(32);
						Authorization.Ready ready = new Authorization.Ready(
								salt);
						receptor.send(ready);

						/* Receive password, authenticate and respond. */
						o = receptor.receive();
						if (o instanceof Authorization.Data) {
							Authorization.Data data = (Authorization.Data) o;
							Authorization.Response response = new Authorization.Response(
									false);

							this.user = Storage.instance().logIn(userName,
									data.getHashedPass(), salt);

							if (user != null) {
								response = new Authorization.Response(true);
								this.authorized = true;
								UserList.instance().logUser(userName,
										ClientHandler.this);
								receptor.send(response);
								logger.log(Level.INFO,
										"Client user {0} authorized from {1}.",
										new Object[] {
												userName,
												receptor.getInetAddress()
														.getHostName() });
							} else {
								receptor.send(response);
								break;
							}
						}
					}

					/* -- Requests not requiring a persistent Project -- */

					/* Handle list projects request. */
					else if (o instanceof ListProjectsRequest) {
						if (!authorized)
							break;

						String[] list = user.listProjects();
						ListProjectsResponse data = new ListProjectsResponse(
								list);
						receptor.send(data);
					}

					/* Monitoring projects information request. */
					else if (o instanceof MonitorTasksRequest) {
						if (!authorized)
							break;

						String[] projectList = user.listProjects();
						int count = projectList.length;
						MonitorTasksResponse response = new MonitorTasksResponse(
								count);
						receptor.send(response);

						for (int i = 0; i < projectList.length; i++) {
							Project project = user.getProject(projectList[i]);
							int complete = project.countRenderedFrames();
							int total = project.countTotalFrames();
							MonitorItem item = new MonitorItem(projectList[i],
									total, complete, false, 0);
							receptor.send(item);
						}
					}

					/* Downloading frames. */
					else if (o instanceof FramesRequest) {
						if (!authorized)
							break;

						FramesRequest request = (FramesRequest) o;
						String projectName = request.getProjectName();
						Project project = user.getProject(projectName);

						FramesResponse response;
						if (project == null) {
							response = new FramesResponse(0);
							receptor.send(response);
							continue;
						}

						FileMap[] fileMaps;
						fileMaps = project.getFrameFileMaps();
						int count = fileMaps.length;
						response = new FramesResponse(count);
						receptor.send(response);

						LocalPath path;
						File file;
						long size;
						FileTransfer.Header header;
						BufferedInputStream in;
						for (int i = 0; i < count; i++) {
							path = fileMaps[i].getDestinationPath();
							file = fileMaps[i].getSourcePath();
							size = file.length();
							in = new BufferedInputStream(new FileInputStream(
									file));
							header = new FileTransfer.Header(path, size);
							receptor.send(header);
							FileTransfer.transferData(in, receptor.out, size);
							in.close();
						}
					}

					/* Delete a project. */
					else if (o instanceof DeleteRequest) {
						if (!authorized)
							break;

						DeleteRequest request = (DeleteRequest) o;
						String projectName = request.getProjectName();

						try {
							user.deleteProject(projectName);
						} catch (StorageException e) {
							receptor.send(new DeleteResponse(false));
							continue;
						}
						receptor.send(new DeleteResponse(true));
					}

					/* -- Requests making use of the persistent Project -- */

					/* Create new project request. */
					else if (o instanceof CreateProjectRequest) {
						if (!authorized)
							break;

						/* Receive project name and check uniqueness. */
						CreateProjectRequest request = (CreateProjectRequest) o;
						String projectName = request.getProjectName();
						boolean exists = user.hasProject(projectName);

						/* Make the project. */
						boolean created = false;
						if (!exists) {
							try {
								user.makeProject(projectName);
								created = true;
							} catch (StorageException e) {
								created = false;
								logger.log(Level.SEVERE,
										"Failed to make new project, {0}.",
										projectName);
							}
						}

						currentProject = user.getProject(projectName);

						/* Reply with uniqueness. */
						CreateProjectResponse response = new CreateProjectResponse(
								created);
						receptor.send(response);
					}

					/* Request to add files to a project. */
					else if (o instanceof SubmitFilesRequest) {
						if (!authorized || currentProject == null)
							break;

						SubmitFilesRequest request = (SubmitFilesRequest) o;
						int fileCount = request.getFileCount();
						Archive archive = currentProject.makeArchiveForUpload();

						/*
						 * Receive files into one big (temporary) file and store
						 * headers in a separate index file.
						 */
						BufferedOutputStream dataOut = new BufferedOutputStream(
								new FileOutputStream(archive.getTempDataFile()));
						ObjectOutputStream indexOut = new ObjectOutputStream(
								new BufferedOutputStream(new FileOutputStream(
										archive.getIndexFile())));

						long size, transferred;
						for (int i = 0; i < fileCount; i++) {
							FileTransfer.Header header = (FileTransfer.Header) receptor
									.receive();
							size = header.getSize();
							indexOut.writeUnshared(header);
							transferred = FileTransfer.transferData(
									receptor.in, dataOut, size);
							if (transferred != size) {
								logger
										.severe("File transfer error. Mismatched sizes.");
								archive.delete();
								receptor.close();
								return;
							}
						}
						dataOut.close();
						indexOut.close();

						/* Make temporary data file the final data file. */
						boolean b = archive.finalizeUpload();
						if (!b)
							logger.severe("Failed to finalize upload.");

						/* Send final response. */
						SubmitFilesResponse response = new SubmitFilesResponse(
								b);
						receptor.send(response);
					}

					/* Request to add tasks to a project. */
					else if (o instanceof SubmitTasksRequest) {
						if (!authorized || currentProject == null)
							break;

						SubmitTasksRequest request = (SubmitTasksRequest) o;
						boolean success = true;

						/* Add the tasks. */
						try {
							for (Task task : request.getTasks())
								currentProject.addTask(task);
						} catch (StorageException e) {
							logger.severe("Error adding tasks.");
							logger.severe(e.getLocalizedMessage());
							success = false;
						}

						/* Send response. */
						SubmitTasksResponse response = new SubmitTasksResponse();
						receptor.send(response);
					}

				} catch (IOException e) {
					break;
				}
			}
			receptor.close();
		}

		/**
		 * 
		 */
		private boolean authorized = false;

		/**
		 * 
		 */
		private User user = null;
	}
}
