/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007-2008: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.distributor;

import java.awt.AWTException;
import java.awt.GraphicsEnvironment;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.distributor.gui.TrayController;
import net.sf.distriblend.distributor.gui.Wizard;
import net.sf.distriblend.distributor.storage.Storage;
import net.sf.distriblend.distributor.storage.StorageException;
import net.sf.distriblend.share.Settings;
import net.sf.distriblend.share.net.Server;

/**
 * The entry point class for the distributor application. The role of the
 * distributor is to receive files and tasks and to distribute these resources
 * to connected nodes as required.
 * 
 * <p>
 * The {@link Storage} class dictates how files are stored by the distributor.
 * 
 * @author Ian Thompson
 * 
 */
public class Main {

	/**
	 * The main entry point to the distributor application.
	 * <p>
	 * There are three command line arguments, to create and load a settings
	 * file and to create users:
	 * </p>
	 * <table border="1">
	 * <tr>
	 * <td><code>-s</code> <em>distributor.conf</em></td>
	 * <td>Loads the settings from the file <em>distributor.conf</em>.
	 * Missing settings will be given their default values.</td>
	 * </tr>
	 * <tr>
	 * <td><code>-t</code> <em>distributor.conf</em></td>
	 * <td>Creates a template settings file and saves it to
	 * <em>distributor.conf</em>. This file will contain the application
	 * default settings.</td>
	 * </tr>
	 * <tr>
	 * <td><code>-u</code> <em>username</em> <em>password</em></td>
	 * <td>Creates the user <em>username</em> with the password
	 * <em>password</em>.</td>
	 * </tr>
	 * </table>
	 * <p>
	 * If no settings file is specified the following directories will be
	 * searched for a <code>distributor.conf</code> settings file:
	 * <ul>
	 * <li><code>System.getProperty("user.dir");</code>
	 * <li><code>System.getProperty("user.home");</code>
	 * <li><code>System.getProperty("user.home") + File.separator + ".distriblend";</code>
	 * </ul>
	 * Failing that, an error will be logged and the program will terminate.
	 * </p>
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		new Main(args);
	}

	/**
	 * Creates a distributor application with the given command line arguments.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	private Main(String[] args) {
		logger.info("Distributor starting...");

		/* Add a shutdown hook. */
		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			public void run() {
				logger.info("Distributor terminated.");
			}
		}));

		/* Add tray icon interface */
		TrayController controller = null;
		try {
			controller = new TrayController(this);
		} catch (AWTException e) {
			logger.warning("System tray unavailable, using console only.");
		}

		/* Set default settings (all used settings must have a default). */
		Settings settings = Settings.instance();

		settings.setStringProperty("storage.root", "storage" + File.separator
				+ "distributor");
		settings.setDescription("storage.root", "Specifies the directory to "
				+ "use for the storage of project files, tasks and frame "
				+ "files. Everything is written to and read from this "
				+ "directory and its sub-directories.");

		settings.setStringProperty("server.client.bind", "");
		settings.setDescription("server.client.bind", "The address to which "
				+ "the client server socket should be bound. Leave this "
				+ "blank to bind on all address/interfaces.");

		settings.setIntProperty("server.client.port", 55100);
		settings.setDescription("server.client.port", "The port on which to "
				+ "listen for incoming client connections.");

		settings.setIntProperty("server.client.timeout", 3600000);
		settings.setDescription("server.client.timeout", "How long to wait "
				+ "when reading on a client connection before timing out.");

		settings.setStringProperty("server.node.bind", "");
		settings.setDescription("server.node.bind", "The address to which "
				+ "the node server socket should be bound. Leave this blank "
				+ "to bind on all address/interfaces.");

		settings.setIntProperty("server.node.port", 55101);
		settings.setDescription("server.node.port", "The port on which to "
				+ "listen for incoming node connections.");

		settings.setIntProperty("server.node.timeout", 30000);
		settings.setDescription("server.node.timeout", "How long to wait "
				+ "when reading on a node connection before timing out.");

		settings.setStringProperty("server.node.user", "node");
		settings.setDescription("server.node.user", "The user name used "
				+ "when authorizing connections from a node.");

		settings.setStringProperty("server.node.password", "password");
		settings.setDescription("server.node.password", "The password to use "
				+ "when authorizing connections from a node.");

		/* Parse command line. */
		boolean showUsage = false;
		boolean settingsLoaded = false;
		boolean addUser = false;
		String username = null, password = null;

		if (args.length > 0) {
			int i = 0;
			String option;
			String value;
			File file;
			whileLabel: while (i < args.length) {
				option = args[i++].trim();
				if (option.length() < 2 || option.charAt(0) != '-') {
					showUsage = true;
					break;
				}
				char opt = option.charAt(1);
				try {
					switch (opt) {
					case 't':
						value = args[i++].trim();
						file = new File(value);
						logger.info("Creating template settings file: \""
								+ file.getAbsolutePath() + "\"...");
						settings.save(file);
						logger.info("Settings saved.");

						/* Remove tray icon before quitting. */
						if (controller != null)
							controller.remove();
						return;

					case 's':
						value = args[i++].trim();
						file = new File(value);
						logger.info("Loading settings file: \""
								+ file.getAbsolutePath() + "\"...");
						settings.load(file);
						logger.info("Settings loaded.");
						settingsLoaded = true;
						break;

					case 'u':
						username = args[i++].trim();
						password = args[i++].trim();
						addUser = true;
						break;

					case 'h':
					case '?':
						showUsage = true;
						break whileLabel;

					default:
						showUsage = true;
						break whileLabel;
					}
				} catch (IOException e) {
					logger.severe(e.getLocalizedMessage());
					showUsage = true;
					break whileLabel;
				} catch (IndexOutOfBoundsException e) {
					logger.severe("Wrong number of arguments.");
					showUsage = true;
					break whileLabel;
				}
			}
		}

		/* Attempt to load settings from common locations. */
		if (!settingsLoaded) {
			String settingsName = "distributor.conf";
			String workingDir = System.getProperty("user.dir");
			String homeDir = System.getProperty("user.home");
			File file;

			file = new File(workingDir, settingsName);
			try {
				settings.load(file);
				settingsLoaded = true;
			} catch (IOException e1) {
				file = new File(homeDir, settingsName);
				try {
					settings.load(file);
					settingsLoaded = true;
				} catch (IOException e2) {
					file = new File(homeDir + File.separator + ".distriblend",
							settingsName);
					try {
						settings.load(file);
						settingsLoaded = true;
					} catch (IOException e3) {
						logger.severe("No settings file found. Use the "
								+ "configuration wizard to create one or use "
								+ "the -t option to create a template.");
					}
				}
			}
		}

		/* Only allow distributor to be run with a settings file. */
		if (!settingsLoaded) {
			if (!GraphicsEnvironment.isHeadless()) {
				logger.info("Loading Configuration Wizard...");
				int result = Wizard.showConfigurationDialog();
				if (result == Wizard.FINISH_OPTION)
					settingsLoaded = true;
				else
					showUsage = true;
			} else
				showUsage = true;
		}

		/* If requested, display usage and quit. */
		if (showUsage) {
			System.out.println("Usage: distributor [-s|-t distributor.conf]");
			System.out.println();
			System.out.println(" -s distributor.conf   Loads settings from "
					+ "the file distributor.conf.");
			System.out.println();
			System.out.println(" -t distributor.conf   Creates a template "
					+ "settings file distributor.conf and");
			System.out.println("                       exits.");
			System.out.println();
			System.out.println(" -u username password  Adds the user username "
					+ "to the stored list with the");
			System.out.println("                       password password. A "
					+ "settings file must be loaded");
			System.out.println("                       implicitly or "
					+ "explicitly before using this option.");
			System.out.println();
			System.out.println("If no file is loaded with -s then the current "
					+ "directory, the user's home");
			System.out.println("directory and the .distriblend sub-directory "
					+ "(of the user's home directory)");
			System.out.println("will be searched for distributor.conf");

			/* Remove tray icon before quitting. */
			if (controller != null)
				controller.remove();
			return;
		}

		/* Initialize storage system. */
		try {
			Storage.initialize();
		} catch (StorageException e) {
			logger
					.severe("Cannot create storage directory (see console for details).");
			logger.severe(e.getLocalizedMessage());
			System.out.println();
			System.out.println("Please create the storage directory as "
					+ "instructed in the Getting Started document.");
			System.out.println();
			return;
		}

		/* Create a user if requested. */
		if (addUser) {
			try {
				Storage.instance().createUser(username, password);
			} catch (StorageException e) {
				logger.log(Level.SEVERE, "Failed to add user.", e);
			}
			logger.log(Level.INFO, "User {0} added.", username);
			if (controller != null)
				controller.remove();
			return;
		}

		/* Start the servers. */
		startServers();
	}

	/**
	 * 
	 * 
	 */
	public void shutdown() {
		logger.info("Shutting down...");
		clientServer.shutdown();
		nodeServer.shutdown();
	}

	/**
	 * Restart the servers.
	 * 
	 */
	public void restart() {
		shutdown();
		startServers();
	}

	/**
	 * 
	 */
	private static Logger logger = Logger.getLogger(Main.class.getName());

	/**
	 * 
	 */
	private Server clientServer;

	/**
	 * 
	 */
	private Server nodeServer;

	/**
	 * Starts the servers.
	 */
	private void startServers() {
		int port = Settings.instance().getIntProperty("server.client.port");
		clientServer = new Server(port, ClientHandler.class);
		logger.log(Level.INFO, "Listening for client connections on port "
				+ "{0,Number,#}.", port);

		port = Settings.instance().getIntProperty("server.node.port");
		nodeServer = new Server(port, NodeConnectionHandler.class);
		logger.log(Level.INFO, "Listening for node connections on port "
				+ "{0,Number,#}.", port);

		logger.info("Distributor started.");
	}
}
