/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.distributor;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.distributor.storage.Project;
import net.sf.distriblend.distributor.storage.Storage;
import net.sf.distriblend.distributor.storage.User;
import net.sf.distriblend.distributor.storage.Project.Archive;
import net.sf.distriblend.share.HashFunction;
import net.sf.distriblend.share.LocalPath;
import net.sf.distriblend.share.Settings;
import net.sf.distriblend.share.Task;
import net.sf.distriblend.share.TaskFrame;
import net.sf.distriblend.share.net.ConnectionHandler;
import net.sf.distriblend.share.net.Receptor;
import net.sf.distriblend.share.protocol.Authorization;
import net.sf.distriblend.share.protocol.FileTransfer;
import net.sf.distriblend.share.protocol.Distribution.FilesRequest;
import net.sf.distriblend.share.protocol.Distribution.FilesResponse;
import net.sf.distriblend.share.protocol.Distribution.FrameUpload;
import net.sf.distriblend.share.protocol.Distribution.TaskFrameState;
import net.sf.distriblend.share.protocol.Distribution.TaskRequest;
import net.sf.distriblend.share.protocol.Distribution.TaskResponse;
import net.sf.distriblend.share.protocol.Distribution.TaskUpdate;
import net.sf.distriblend.share.protocol.Distribution.UpdateReceived;
import net.sf.distriblend.share.protocol.Distribution.UploadReceived;
import net.sf.distriblend.share.protocol.FileTransfer.Header;

/**
 * @author Ian Thompson
 * 
 */
public class NodeConnectionHandler extends ConnectionHandler {

	/**
	 * 
	 */
	public NodeConnectionHandler() {
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void handleConnection(Socket socket) throws IOException {
		logger.log(Level.INFO, "Node connected from {0}.", socket
				.getInetAddress().getHostName());
		if (thread == null) {
			int timeout = Settings.instance().getIntProperty(
					"server.node.timeout");
			this.receptor = new Receptor(socket, timeout);
			this.thread = new NodeHandlerThread();
			this.thread.start();
		}
	}

	/**
	 * Forcibly disconnects the client.
	 */
	public void disconnect() {
		receptor.close();
	}

	/* -- Private -- */

	/**
	 * 
	 */
	private static Logger logger = Logger.getLogger(NodeConnectionHandler.class
			.getName());

	/**
	 * 
	 */
	private NodeHandlerThread thread = null;

	/**
	 * 
	 */
	private Receptor receptor;

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class NodeHandlerThread extends Thread {

		/**
		 * 
		 * @throws IOException
		 */
		public NodeHandlerThread() throws IOException {
			super("NodeHandlerThread");
			this.setDaemon(true);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			while (true) {
				try {
					Object o = receptor.receive();

					/* Authorize request. */
					if (o instanceof Authorization.Request) {
						Authorization.Request request = (Authorization.Request) o;
						String userName = request.getUser();

						/* Send salt for encoding password. */
						String salt = Authorization.generateSalt(32);
						String serverUser = Settings.instance()
								.getStringProperty("server.node.user");
						String serverHashedPassword;
						try {
							serverHashedPassword = HashFunction
									.getDefault()
									.hash(
											Settings
													.instance()
													.getStringProperty(
															"server.node.password")
													+ salt);
						} catch (NoSuchAlgorithmException e) {
							logger.severe("Hash algorithm not installed.");
							break;
						}
						Authorization.Ready ready = new Authorization.Ready(
								salt);
						receptor.send(ready);

						/* Receive password, authenticate and respond. */
						o = receptor.receive();
						if (o instanceof Authorization.Data) {
							Authorization.Data data = (Authorization.Data) o;
							Authorization.Response response = new Authorization.Response(
									false);

							if (userName.equals(serverUser)
									&& data.getHashedPass().equals(
											serverHashedPassword)) {
								response = new Authorization.Response(true);
								this.authorized = true;
								receptor.send(response);
								logger.log(Level.INFO,
										"Node user {0} authorized from {1}.",
										new Object[] {
												userName,
												receptor.getInetAddress()
														.getHostName() });
							} else {
								receptor.send(response);
								break;
							}
						}
					}

					/* Ready for task request */
					else if (o instanceof TaskRequest) {
						if (!authorized || currentTaskFrame != null)
							break;

						TaskFrame taskFrame = TaskQueue.instance().next();
						if (taskFrame == null) {
							receptor.send(new TaskResponse(null, 5));
							continue;
						}

						/* Send task and frame to node. */
						Task task = taskFrame.getTask();
						User user = Storage.instance().getUser(task.getUser());
						Project project = user
								.getProject(task.getProjectName());
						int archiveCount = project.getArchiveCount();
						TaskResponse response = new TaskResponse(taskFrame,
								archiveCount);
						receptor.send(response);

						currentTaskFrame = taskFrame;
					}

					/* File dependency request. */
					else if (o instanceof FilesRequest) {
						if (!authorized || currentTaskFrame == null)
							break;

						FilesRequest request = (FilesRequest) o;
						int possessed = request.getArchivesPossessed();
						Task task = currentTaskFrame.getTask();
						User user = Storage.instance().getUser(task.getUser());
						Project project = user
								.getProject(task.getProjectName());
						int needed = project.getArchiveCount();

						/* Inform the node how many archives to receive. */
						FilesResponse response = new FilesResponse(needed
								- possessed);
						receptor.send(response);

						/* Transfer the missing archives. */
						long size;
						Archive archive;
						File file;
						BufferedInputStream in;
						LocalPath path;
						for (int i = possessed; i < needed; i++) {
							archive = project.getArchive(i);

							/* Send the index file. */
							file = archive.getIndexFile();
							path = new LocalPath(file.getName(), false);
							size = file.length();
							receptor.send(new FileTransfer.Header(path, size));
							in = new BufferedInputStream(new FileInputStream(
									file));
							FileTransfer.transferData(in, receptor.out, size);

							/* Send the data file. */
							file = archive.getDataFile();
							path = new LocalPath(file.getName(), false);
							size = file.length();
							receptor.send(new FileTransfer.Header(path, size));
							in = new BufferedInputStream(new FileInputStream(
									file));
							FileTransfer.transferData(in, receptor.out, size);
						}
					}

					/* Task update. */
					else if (o instanceof TaskUpdate) {
						if (!authorized || currentTaskFrame == null)
							break;

						TaskUpdate update = (TaskUpdate) o;
						TaskFrameState state = update.getState();
						int frame = currentTaskFrame.getFrame();
						InetAddress address = receptor.getInetAddress();
						switch (state) {
						case STARTED:
							logger.log(Level.INFO,
									"Rendering frame {0} on {1}.",
									new Object[] { frame, address });
							break;
						case FAILED:
							TaskQueue.instance()
									.finish(currentTaskFrame, false);
							logger.log(Level.WARNING,
									"Rendering failed on {0} rendering "
											+ "frame {1}.", new Object[] {
											address, frame });
							currentTaskFrame = null;
							break;
						case COMPLETE:
							logger.log(Level.INFO,
									"Rendering completed on {0} rendering "
											+ "frame {1}.", new Object[] {
											address, frame });
							break;
						}

						receptor.send(new UpdateReceived());
					}

					/* Frame file upload. */
					else if (o instanceof FrameUpload) {
						if (!authorized || currentTaskFrame == null)
							break;

						Task task = currentTaskFrame.getTask();
						int frame = currentTaskFrame.getFrame();

						User user = Storage.instance().getUser(task.getUser());
						Project project = user
								.getProject(task.getProjectName());

						/* Receive the frame file. */
						Header header = (Header) receptor.receive();
						File file = project.frameFile(task, frame);
						file.getParentFile().mkdirs();
						BufferedOutputStream out = new BufferedOutputStream(
								new FileOutputStream(file));
						FileTransfer.transferData(receptor.in, out, header
								.getSize());
						out.close();

						receptor.send(new UploadReceived());
						TaskQueue.instance().finish(currentTaskFrame, true);
						currentTaskFrame = null;
					}

				} catch (IOException e) {
					/* Connection is broken, put frame back on the queue. */
					if (currentTaskFrame != null)
						TaskQueue.instance().cancel(currentTaskFrame);
					break;
				}
			}
			receptor.close();
		}

		private boolean authorized = false;

		/** Used to persist a single task across multiple commands. */
		private TaskFrame currentTaskFrame = null;
	}
}
