/*
 * Distriblend: A distributed rendering system for Blender Copyright (C) 2007:
 * Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.distributor.storage;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.security.NoSuchAlgorithmException;
import java.util.Collections;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.share.SafeString;
import net.sf.distriblend.share.Settings;

/**
 * <p>
 * Presents a system of user, project and task records that are mirrored by a
 * file structure on disk to maintain persistency. When first accessed (by the
 * {@link #instance()} method) this class loads serialized records from disk. As
 * more users, projects and tasks are added they are serialized and saved. This
 * 'read at startup, write while running' paradigm together with synchronization
 * ensures a safe storage mechanism.
 * </p>
 * <p>
 * The storage system uses the following path structures (replacing
 * <code>/</code> with the system file separator {@link File#separator}):
 * <table border="1">
 * <tr>
 * <td><code>{ROOT}</code></td>
 * <td>Root directory. Assumed to hold only sub-directories that represent user
 * names. Taken from the <code>"storage.root"</code> property of
 * {@link Settings}, this directory must allow the creation of sub-directories
 * if new users are to be added.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}</code></td>
 * <td>User's storage directory (where {USER} is replaced by the user's name).
 * This directory must hold a <code>"user.ini"</code> file and be writable by
 * the Distributor application.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/user.ini</code></td>
 * <td>User profile. Holds user password and preference settings.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/projects</code></td>
 * <td>Projects directory. Contains sub-directories for each project.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/projects/{PROJECT}</code></td>
 * <td>Project storage directory. Holds the sub-directories as follows...</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/projects/{PROJECT}/archives</code></td>
 * <td>Project archives directory. Stores uploaded archive files in the form
 * <code>"archive_{NUM}.zip"</code> where {NUM} is a unique number to allow
 * for multiple file sets to be uploaded. The first file set archive will always
 * be <code>"archive_0.zip"</code>. While uploading <code>"_upload"</code>
 * is appended to the file name. This is removed once the upload has
 * successfully completed.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/projects/{PROJECT}/tasks</code></td>
 * <td>Project tasks directory. Stores and maintains a list of tasks that have
 * been submitted and their progresses. Each task is stored in the format
 * <code>"task_{NUM}.task"</code> where {NUM} is a unique number incremented
 * each time a task is added.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/projects/{PROJECT}/frames</code></td>
 * <td>Frame storage directory. Holds sub-directories for each source file,
 * created when frames are first received.</td>
 * </tr>
 * <tr>
 * <td><code>{ROOT}/{USER}/frames/{PROJECT}/frames/{SOURCEFILE}</code></td>
 * <td>Rendered frames directory. Holds the actual frame files received from
 * the nodes network. These files are of the form <code>"{FRMNUM}.{EXT}"</code>,
 * where {FRMNUM} is the formatted frame number and {EXT} is the chosen
 * extension (for example, <code>"0001.png"</code>).</td>
 * </tr>
 * </table>
 * </p>
 * 
 * @author Ian Thompson
 * 
 */
public class Storage {

	/* -- Static methods -- */

	/**
	 * Initializes the storage system. This method make take some time to
	 * complete but must be called before attempting to call {@link #instance()}.
	 * The time this method takes to complete will depend on the number of user
	 * directories, the number of projects and other, platform specific factors.
	 * 
	 * @throws StorageException
	 */
	public static void initialize() throws StorageException {
		if (instance == null) {
			File root = new File(Settings.instance().getStringProperty(
					"storage.root"));
			root.mkdirs();
			try {
				instance = new Storage(root, null);
			} catch (FileNotFoundException e) {
				logger.log(Level.SEVERE,
						"Storage root directory must exist: {0}.", root
								.getPath());
			}
		}
	}

	/**
	 * Returns the single {@link Storage} instance.
	 * 
	 * @return The single {@link Storage} instance.
	 */
	public static Storage instance() {
		return instance;
	}

	/* -- Users and authorization -- */

	/**
	 * Attempts to log the specified user into the system. If the user name does
	 * not exist or the password does not match <code>null</code> is returned.
	 * 
	 * @param userName
	 *            The name of the user to log in.
	 * @param hashedPassword
	 *            The password as hashed with given salt.
	 * @param salt
	 *            The salt used when hashing the password.
	 * @return If the user exists and the password matches the user's profile is
	 *         returned. Otherwise <code>null</code> is returned.
	 */
	public User logIn(String userName, String hashedPassword, String salt) {
		User user = users.get(userName);
		try {
			if (user == null || !user.checkHashedPassword(hashedPassword, salt))
				return null;
		} catch (NoSuchAlgorithmException e) {
			logger.severe("Hash algorithm not installed.");
			return null;
		}

		return user;
	}

	/**
	 * Attempts to find and return the specified user on the system. If no user
	 * is found <code>null</code> is returned.
	 * 
	 * @param userName
	 *            The name of the user.
	 * @return The located user or <code>null</code> if none was found.
	 */
	public User getUser(String userName) {
		return users.get(userName);
	}

	/**
	 * Returns the list of user names held by this storage system.
	 * 
	 * @return A list of user names.
	 */
	public String[] listUsers() {
		String[] userNames;
		synchronized (users) {
			userNames = new String[users.size()];
			int i = 0;
			for (String name : users.keySet())
				userNames[i++] = name;
		}
		return userNames;
	}

	/* -- Private -- */

	/** The single instance of this class. */
	private static Storage instance = null;

	/** The logger for this class. */
	private static Logger logger = Logger.getLogger(Storage.class.getName());

	/** Storage root directory (ROOT). */
	private final File root;

	/** Users registered with the system, at startup and when added. */
	private final Map<String, User> users;

	/**
	 * 
	 * @param root
	 * @param sub
	 * @throws IllegalArgumentException
	 * @throws FileNotFoundException
	 */
	private Storage(File root, File sub) throws IllegalArgumentException,
			FileNotFoundException {

		if (root == null)
			throw new FileNotFoundException("Root cannot be null.");
		if (!root.exists())
			throw new FileNotFoundException("Root must exist.");
		if (sub != null && sub.isAbsolute())
			throw new IllegalArgumentException(
					"The secondary path component must be relative.");

		this.root = root;

		/* Load records from disk. */
		logger.log(Level.INFO, "Parsing storage directories: \"{0}\".",
				this.root);
		this.users = Collections.synchronizedMap(new TreeMap<String, User>());

		for (File userHome : this.root.listFiles()) {
			if (userHome.isDirectory()) {
				String user = userHome.getName();

				try {
					logger.log(Level.FINE, "Adding user {0}.", user);
					User p = new User(user, userHome);
					users.put(user, p);
				} catch (IOException e) {
					logger.log(Level.SEVERE,
							"Problem reading user {0}'s profile.", user);
					logger.severe(e.getLocalizedMessage());
				} catch (StorageException e) {
					logger.log(Level.SEVERE, e.getLocalizedMessage());
				}
			}
		}
	}

	/**
	 * @param username
	 * @param password
	 * @throws StorageException
	 */
	public void createUser(String username, String password)
			throws StorageException {
		if (!SafeString.isSafe(username))
			throw new StorageException("Username invalid.");
		File home = new File(root, username);
		home.mkdirs();
		File file = new File(home, "user.ini");
		PrintWriter out;
		try {
			out = new PrintWriter(new BufferedWriter(new FileWriter(file)));
			out.println(password);
			out.close();
		} catch (IOException e) {
			throw new StorageException("Failed to write user.ini");
		}
	}
}
