/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007-2008: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.node;

import java.awt.AWTException;
import java.awt.GraphicsEnvironment;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.node.gui.TrayController;
import net.sf.distriblend.node.gui.Wizard;
import net.sf.distriblend.share.LocalPath;
import net.sf.distriblend.share.Settings;
import net.sf.distriblend.share.Task;
import net.sf.distriblend.share.TaskFrame;
import net.sf.distriblend.share.net.AuthorizedConnector;
import net.sf.distriblend.share.net.event.ConnectionEvent;
import net.sf.distriblend.share.net.event.ConnectionListener;
import net.sf.distriblend.share.protocol.FileTransfer;
import net.sf.distriblend.share.protocol.Distribution.FilesRequest;
import net.sf.distriblend.share.protocol.Distribution.FilesResponse;
import net.sf.distriblend.share.protocol.Distribution.FrameUpload;
import net.sf.distriblend.share.protocol.Distribution.TaskFrameState;
import net.sf.distriblend.share.protocol.Distribution.TaskRequest;
import net.sf.distriblend.share.protocol.Distribution.TaskResponse;
import net.sf.distriblend.share.protocol.Distribution.TaskUpdate;
import net.sf.distriblend.share.protocol.FileTransfer.Header;

/**
 * The entry point class for the node application. The main role of the node is
 * to render tasks. To achieve this, it connects to a distributor and requests a
 * task and the files required to execute this task. Once complete the rendered
 * frames are returned and the next task requested.
 * <p>
 * The {@link Directories} class dictates how files are stored by the node.
 * 
 * @author Ian Thompson
 * 
 * @see Directories
 * @see Settings
 */
public class Main implements ConnectionListener {

	/**
	 * The main entry point to the node application.
	 * <p>
	 * There are currently only two command line arguments, to create and load a
	 * settings file:
	 * </p>
	 * <table border="1">
	 * <tr>
	 * <td><code>-t</code> <em>node.conf</em></td>
	 * <td>Creates a template settings file and saves it to <em>node.conf</em>.
	 * This file will contain the application default settings.</td>
	 * </tr>
	 * <tr>
	 * <td><code>-s</code> <em>node.conf</em></td>
	 * <td>Loads the settings from the file <em>node.conf</em>. Missing
	 * settings will be given their default values.</td>
	 * </tr>
	 * </table>
	 * <p>
	 * If no settings file is specified the following directories will be
	 * searched for a <code>node.conf</code> settings file:
	 * <ul>
	 * <li><code>System.getProperty("user.dir");</code>
	 * <li><code>System.getProperty("user.home");</code>
	 * <li><code>System.getProperty("user.home") + File.separator + ".distriblend";</code>
	 * </ul>
	 * Failing that, an error will be logged and the program will terminate.
	 * </p>
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		new Main(args);
	}

	/**
	 * Shuts down each server.
	 */
	public void shutdown() {
		logger.info("Shutting down...");
		if (running) {
			running = false;
			connector.close();
		}
	}

	/* -- Private -- */

	/**
	 * The {@link Logger} for this class.
	 */
	private static Logger logger = Logger.getLogger(Main.class.getName());

	/**
	 * Creates a node application with the given command line arguments.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	private Main(String[] args) {
		logger.info("Node starting...");

		/* Add a shutdown hook. */
		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			public void run() {
				logger.info("Node terminated.");
			}
		}));

		/* Add tray icon interface */
		TrayController controller = null;
		try {
			controller = new TrayController(this);
		} catch (AWTException e) {
			logger.warning("System tray unavailable, using console only.");
		}

		/* Set default settings (all used settings must have a default). */
		Settings settings = Settings.instance();

		settings.setStringProperty("storage.root", "storage" + File.separator
				+ "node");
		settings.setDescription("storage.root", "Specifies the directory to "
				+ "use for the storage of project files, tasks and frame "
				+ "files. Everything is written to and read from this "
				+ "directory and its sub-directories.");

		settings.setStringProperty("connect.distributor.host", "");
		settings.setDescription("connect.distributor.host", "The address of "
				+ "the distributor host machine. This must be a valid IP "
				+ "address or host name.");

		settings.setIntProperty("connect.distributor.port", 55101);
		settings.setDescription("connect.distributor.port", "The port to "
				+ "connect to on the distributor. The distributor must have "
				+ "the same value for its server.node.port setting.");

		settings.setIntProperty("connect.distributor.timeout", 30000);
		settings.setDescription("connect.distributor.timeout", "When "
				+ "communicating with the distributor, this setting dictates "
				+ "how long to wait when reading responses before timing out.");

		settings.setStringProperty("connect.distributor.user", "node");
		settings.setDescription("connect.distributor.user", "The user name "
				+ "to use when connecting to the distributor.");

		settings.setStringProperty("connect.distributor.password", "password");
		settings.setDescription("connect.distributor.password", "The password "
				+ "to use when connecting to the distributor.");

		settings.setStringProperty("blender.command", "blender");
		settings.setDescription("blender.command", "The command used to "
				+ "execute Blender on the local machine. If Blender is not in "
				+ "the system path then you will need to set this to the "
				+ "path of the Blender executable. On Windows this is usually "
				+ "C:\\Program Files\\Blender Foundation\\Blender\\blender"
				+ ".exe");

		/* Parse command line. */
		boolean showUsage = false;
		boolean settingsLoaded = false;

		if (args.length > 0) {
			int i = 0;
			String option;
			String value;
			File file;
			whileLabel: while (i < args.length) {
				option = args[i++].trim();
				if (option.length() < 2 || option.charAt(0) != '-') {
					showUsage = true;
					break;
				}
				char opt = option.charAt(1);
				try {
					switch (opt) {
					case 't':
						value = args[i++].trim();
						file = new File(value);
						logger.info("Creating template settings file: \""
								+ file.getAbsolutePath() + "\"...");
						settings.save(file);
						logger.info("Settings saved.");

						/* Remove tray icon before quitting. */
						if (controller != null)
							controller.remove();
						return;

					case 's':
						value = args[i++].trim();
						file = new File(value);
						logger.info("Loading settings file: \""
								+ file.getAbsolutePath() + "\"...");
						settings.load(file);
						logger.info("Settings loaded.");
						settingsLoaded = true;
						break;

					case 'h':
					case '?':
						showUsage = true;
						break whileLabel;

					default:
						showUsage = true;
						break whileLabel;
					}
				} catch (IOException e) {
					logger.severe(e.getLocalizedMessage());
					showUsage = true;
					break whileLabel;
				} catch (IndexOutOfBoundsException e) {
					logger.severe("Wrong number of arguments.");
					showUsage = true;
					break whileLabel;
				}
			}
		}

		/* Attempt to load settings from common locations. */
		if (!settingsLoaded) {
			String settingsName = "node.conf";
			String workingDir = System.getProperty("user.dir");
			String homeDir = System.getProperty("user.home");
			File file;

			file = new File(workingDir, settingsName);
			try {
				settings.load(file);
				settingsLoaded = true;
			} catch (IOException e1) {
				file = new File(homeDir, settingsName);
				try {
					settings.load(file);
					settingsLoaded = true;
				} catch (IOException e2) {
					file = new File(homeDir + File.separator + ".distriblend",
							settingsName);
					try {
						settings.load(file);
						settingsLoaded = true;
					} catch (IOException e3) {
						logger.severe("No settings file found. Use the "
								+ "configuration wizard to create one or use "
								+ "the -t option to create a template.");
					}
				}
			}
		}

		/* Only allow node to be run with a settings file. */
		if (!settingsLoaded) {
			if (!GraphicsEnvironment.isHeadless()) {
				logger.info("Loading Configuration Wizard...");
				int result = Wizard.showConfigurationDialog();
				if (result == Wizard.FINISH_OPTION)
					settingsLoaded = true;
				else
					showUsage = true;
			} else
				showUsage = true;
		}

		/* If requested, display usage and quit. */
		if (showUsage) {
			System.out.println("Usage: node [-s|-t node.conf]");
			System.out.println();
			System.out.println(" -s node.conf  Loads settings from the "
					+ "file node.conf.");
			System.out.println();
			System.out.println(" -t node.conf  Creates a template "
					+ "settings file node.conf and exits.");
			System.out.println();
			System.out.println("If no file is loaded with -s then the current "
					+ "directory, the user's home");
			System.out.println("directory and the .distriblend sub-directory "
					+ "(of the user's home directory)");
			System.out.println("will be searched for node.conf");
			System.out.println();

			/* Remove tray icon before quitting. */
			if (controller != null)
				controller.remove();
			return;
		}

		/* Initialize renderer. */
		if (!Renderer.instance().test() && controller != null) {
			controller.remove();
			logger.info("Check the blender.command in your node configuration "
					+ "file points to the Blender executable.");
			return;
		}

		/* Start the system. */
		logger.info("Node started.");

		/* Run the node life cycle */
		mainLoop();
	}

	private AuthorizedConnector connector = null;

	private boolean running = false;

	/**
	 * Maintains a connection to the distributor and requests tasks and files
	 * for processing.
	 */
	private void mainLoop() {
		String host = Settings.instance().getStringProperty(
				"connect.distributor.host");
		int port = Settings.instance().getIntProperty(
				"connect.distributor.port");
		int timeout = Settings.instance().getIntProperty(
				"connect.distributor.timeout");
		String nodeuser = Settings.instance().getStringProperty(
				"connect.distributor.user");
		String password = Settings.instance().getStringProperty(
				"connect.distributor.password");
		connector = new AuthorizedConnector(host, port, timeout);
		connector.addConnectionListener(this);

		Task lastTask = null;

		running = true;
		while (running) {

			if (connector.isDisconnected()) {

				/* Start connect thread and wait for it. */
				synchronized (this) {
					try {
						Thread.sleep(1000);
						logger.log(Level.INFO, "Connecting to distributor "
								+ "at {0} port {1}...", new Object[] { host,
								port });
						connector.connect();
						wait();
					} catch (InterruptedException e) {
						return;
					}
				}

				/* If not connected, sleep for a bit then try again. */
				if (!connector.isSuccessfullyConnected()) {
					logger.warning("Connect failed. Retrying in 5 seconds...");
					try {
						Thread.sleep(5000);
					} catch (InterruptedException e) {
						return;
					}
					continue;
				}
			}

			/* Authorize connection */
			if (!connector.authorize(nodeuser, password)) {
				logger.warning("Authorization failed. Retrying in 5 "
						+ "seconds...");
				connector.close();
				try {
					Thread.sleep(5000);
				} catch (InterruptedException e) {
					return;
				}
				continue;
			}

			logger.info("Connected to distributor.");

			try {

				/* Request task */
				TaskResponse taskResponse;
				TaskFrame taskFrame;
				do {
					connector.send(new TaskRequest());
					taskResponse = (TaskResponse) connector.receive();
					taskFrame = taskResponse.getTaskFrame();
					if (taskFrame == null) {
						logger.fine("No tasks available. Trying again in 5 "
								+ "seconds...");
						try {
							Thread.sleep(5000);
							if (!connector.isSuccessfullyConnected())
								return;
						} catch (InterruptedException e) {
							return;
						}
					} else
						break;
				} while (true);

				logger.info("Task received for processing.");

				Task task = taskFrame.getTask();
				String user = task.getUser();
				String projectName = task.getProjectName();

				/*
				 * If this task is from a new project, delete all files from the
				 * old project to save space.
				 */
				if (lastTask == null || !lastTask.getUser().equals(user)
						|| !lastTask.getProjectName().equals(projectName)) {
					Directories.instance().purge();
				}
				lastTask = task;

				/* Request dependencies. */
				int needed = taskResponse.getArchiveCount();
				int possessed = Directories.instance().countArchives(user,
						projectName);
				long size, totalSize, dataRead;
				if (needed > possessed) {
					logger.log(Level.INFO, "Downloading {0} dependencies...",
							needed - possessed);
					connector.send(new FilesRequest(possessed));
					FilesResponse filesResponse = (FilesResponse) connector
							.receive();
					int count = filesResponse.getFileCount();
					Header indexHeader, dataHeader, fileHeader;
					File file;
					BufferedOutputStream fileOut;
					ObjectInputStream indexIn;
					File filesBase = Directories.instance().filesStorage(user,
							projectName);
					InputStream in = connector.getInputStream();

					for (int i = 0; i < count; i++) {

						/* Read the archive index and save to disk. */
						indexHeader = (Header) connector.receive();
						file = Directories.instance().indexFile(user,
								projectName, i);
						file.getParentFile().mkdirs();
						fileOut = new BufferedOutputStream(
								new FileOutputStream(file));
						FileTransfer.transferData(in, fileOut, indexHeader
								.getSize());
						fileOut.close();

						/* Obtain the data header */
						dataHeader = (Header) connector.receive();
						totalSize = dataHeader.getSize();
						dataRead = 0;

						/*
						 * Read back the index file and stream the network data
						 * to its listed files.
						 */
						indexIn = new ObjectInputStream(
								new BufferedInputStream(new FileInputStream(
										file)));
						while (dataRead < totalSize) {
							try {
								fileHeader = (Header) indexIn.readObject();
							} catch (ClassNotFoundException e) {
								logger.severe("Malformed archive index file.");
								throw new IOException();
							}
							file = new File(filesBase, fileHeader.getPath()
									.getPath());
							file.getParentFile().mkdirs();
							size = fileHeader.getSize();
							fileOut = new BufferedOutputStream(
									new FileOutputStream(file));
							FileTransfer.transferData(in, fileOut, size);
							fileOut.close();
							dataRead += size;
						}
					}
				}

				/* Process task. */
				logger.info("Processing task...");
				connector.send(new TaskUpdate(TaskFrameState.STARTED));
				connector.receive();
				Renderer.instance().render(taskFrame);
				File frame = Directories.instance().frameFile(taskFrame);

				if (!frame.exists()) {
					logger.severe("Task failed. Output frame does not exist.");
					connector.send(new TaskUpdate(TaskFrameState.FAILED));
					connector.receive();
					continue;
				}

				logger.info("Task complete.");
				connector.send(new TaskUpdate(TaskFrameState.COMPLETE));
				connector.receive();

				/* Submit rendered frame */
				logger.info("Submitting frame...");
				size = frame.length();
				BufferedInputStream in = new BufferedInputStream(
						new FileInputStream(frame));
				OutputStream out = connector.getOutputStream();
				LocalPath path = new LocalPath(frame.getName(), false);
				connector.send(new FrameUpload());
				connector.send(new Header(path, size));
				FileTransfer.transferData(in, out, size);
				in.close();
				connector.receive();
				logger.info("Frame submitted.");

			} catch (ClassCastException e) {
				logger.log(Level.SEVERE, "Protocol error. Closing "
						+ "connection...", e);
				connector.close();
				continue;

			} catch (IOException e) {
				logger.log(Level.SEVERE, "Communication error. Closing "
						+ "connection...", e);
				connector.close();
				continue;
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void connectionEstablished(ConnectionEvent evt) {
		synchronized (this) {
			notify();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void connectionLost(ConnectionEvent evt) {
		synchronized (this) {
			notify();
		}
	}
}
