/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007-2008: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.node;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.distriblend.share.Settings;
import net.sf.distriblend.share.Task;
import net.sf.distriblend.share.TaskFrame;

/**
 * Presents a static renderer for rendering {@link Task}s in a serial manner.
 * 
 * @author Ian Thompson
 * 
 * @see Task
 */
public class Renderer {

	/**
	 * Returns the static Renderer instance.
	 * 
	 * @return Single renderer instance.
	 */
	public static Renderer instance() {
		if (instance == null) {
			instance = new Renderer();
		}
		return instance;
	}

	/**
	 * Tests the renderer to ensure it is present and functioning.
	 * 
	 * @return Whether the renderer is responsive.
	 */
	public boolean test() {
		String blenderCommand = Settings.instance().getStringProperty(
				"blender.command");
		String[] command = new String[] { blenderCommand, "-v" };
		ProcessBuilder builder = new ProcessBuilder(command);
		builder.redirectErrorStream(true);

		logger.info("Testing renderer...");
		Process blenderProcess = null;
		BufferedReader in = null;
		try {
			blenderProcess = builder.start();
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Failed to start Blender, aborting.", e);
			return false;
		}
		try {
			in = new BufferedReader(new InputStreamReader(blenderProcess
					.getInputStream()));
			String line;
			while ((line = in.readLine()) != null) {
				logger.info("Blender version: " + line);
			}
			in.close();
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Failed to read from Blender.", e);
			blenderProcess.destroy();
			return false;
		}
		int exitCode = -1;
		try {
			exitCode = blenderProcess.waitFor();
		} catch (InterruptedException e) {
		}
		if (exitCode != 0) {
			return false;
		}
		return true;
	}

	/**
	 * Renders the specified task frame and waits for it to complete (or fail)
	 * before returning. The rendered frame, if any, will be saved to the path
	 * generated by {@link Directories#frameFile(TaskFrame)} (for the same
	 * {@link TaskFrame}).
	 * 
	 * @param taskFrame
	 *            The task frame to render.
	 * @return Whether the task was rendered successfully. Note that this does
	 *         not imply that any frames were rendered, or even that the
	 *         renderer executable was even run. It simply states that no errors
	 *         were found in the output generated by the executable.
	 */
	public synchronized boolean render(TaskFrame taskFrame) {
		String blenderCommand = Settings.instance().getStringProperty(
				"blender.command");

		Task task = taskFrame.getTask();
		String user = task.getUser();
		String projectName = task.getProjectName();
		int frame = taskFrame.getFrame();
		String format = task.getImageFormat().codeName;

		File filesStorage = Directories.instance().filesStorage(user,
				projectName);
		File frameStorage = Directories.instance().frameFile(taskFrame)
				.getParentFile();

		File sourceFile = new File(filesStorage, task.getSourceFile().getPath());
		if (!sourceFile.exists())
			return false;

		frameStorage.mkdirs();

		String[] command = new String[] { blenderCommand, "-b",
				sourceFile.getAbsolutePath(), "-o",
				frameStorage.getAbsolutePath() + File.separator, "-F", format,
				"-f", "" + frame };
		// TODO: Add camera/scene script -P preRender.py
		logger.info("Rendering task...");
		logger.log(Level.INFO, "Blender: {0}", Arrays.toString(command));

		ProcessBuilder builder = new ProcessBuilder(command);
		builder.redirectErrorStream(true);
		builder.directory(frameStorage);

		Process blenderProcess = null;
		BufferedReader in = null;
		try {
			blenderProcess = builder.start();
		} catch (IOException e) {
			return false;
		}
		boolean errorLineFound = false;
		try {
			in = new BufferedReader(new InputStreamReader(blenderProcess
					.getInputStream()));
			String line;
			while ((line = in.readLine()) != null) {
				if (line.toUpperCase().indexOf("ERROR") != -1) {
					errorLineFound = true;
					logger.warning("Blender: " + line);
				} else
					logger.fine("Blender: " + line);
			}
			in.close();
		} catch (IOException e) {
			blenderProcess.destroy();
			return false;
		}
		int exitCode = -1;
		try {
			exitCode = blenderProcess.waitFor();
		} catch (InterruptedException e) {
		}
		if (exitCode != 0 || errorLineFound) {
			return false;
		}
		return true;
	}

	/* -- Private -- */

	/**
	 * The single instance of Renderer.
	 */
	private static Renderer instance;

	/**
	 * The {@link Logger} for this class.
	 */
	private static Logger logger = Logger.getLogger(Renderer.class.getName());

	/**
	 * Private constructor to ensure a single instance.
	 */
	private Renderer() {
	}
}
