/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007-2008: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.node.gui;

import java.awt.AWTException;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.TrayIcon;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;

import net.sf.distriblend.node.Main;

/**
 * @author Ian Thompson
 * 
 */
public class TrayController {

	/**
	 * 
	 * @param main
	 * @throws AWTException
	 */
	public TrayController(final Main main) throws AWTException {

		final PopupMenu menu;
		final MenuItem showLogItem;
		final MenuItem shutdownItem;
		final JTextArea textArea;

		if (!SystemTray.isSupported())
			throw new AWTException("System tray is not supported.");

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			logger.severe("Failed to change look and feel.");
		}

		/* Set up logging */
		frame = new JFrame("Node Log");
		frame.setLayout(new BorderLayout());
		frame.setPreferredSize(new Dimension(320, 240));
		textArea = new JTextArea();
		textArea.setEditable(false);
		frame.add(new JScrollPane(textArea), BorderLayout.CENTER);
		frame.add(new JLabel("You may safely close this window, the "
				+ "application will continue running. While the application "
				+ "is running it may be terminated by selecting Shutdown from "
				+ "the tray icon menu."), BorderLayout.SOUTH);
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);

		Handler handler = new Handler() {
			@Override
			public void close() throws SecurityException {
			}

			@Override
			public void flush() {
			}

			@Override
			public void publish(LogRecord record) {
				Formatter formatter = getFormatter();
				if (formatter == null)
					formatter = new SimpleFormatter();
				textArea.append(formatter.format(record));
				textArea.repaint();
			}
		};
		Logger.getLogger("net.sf.distriblend").addHandler(handler);

		/* Set up tray icon */
		final SystemTray tray = SystemTray.getSystemTray();
		Image image = Toolkit.getDefaultToolkit().getImage(
				getClass().getResource("trayicon.png"));
		frame.setIconImage(image);

		/* Set up menu */
		menu = new PopupMenu();
		showLogItem = new MenuItem("Show Log");
		showLogItem.setActionCommand("showLog");
		menu.add(showLogItem);
		shutdownItem = new MenuItem("Shutdown");
		shutdownItem.setActionCommand("shutdown");
		menu.add(shutdownItem);

		trayIcon = new TrayIcon(image, "Distriblend Node", menu);

		ActionListener actionListener = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = e.getActionCommand();
				if (s.equals("shutdown")) {
					try {
						main.shutdown();
						frame.dispose();
						tray.remove(trayIcon);
					} catch (Exception ex) {
						ex.printStackTrace();
						System.exit(-1);
					}
				} else if (s.equals("showLog")) {
					frame.setVisible(true);
				}
			}
		};
		showLogItem.addActionListener(actionListener);
		shutdownItem.addActionListener(actionListener);

		MouseListener mouseListener = new MouseListener() {

			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() > 1)
					frame.setVisible(true);
			}

			public void mouseEntered(MouseEvent e) {
			}

			public void mouseExited(MouseEvent e) {
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {
			}

		};

		trayIcon.addMouseListener(mouseListener);
		trayIcon.setImageAutoSize(true);

		try {
			tray.add(trayIcon);
		} catch (AWTException e) {
			logger.severe("TrayIcon could not be added.");
		}
		frame.setVisible(true);
	}

	/**
	 * Removes the icon from the system tray and disposes of all associated
	 * resources. If the frame is visible, its close operation is set to
	 * DISPOSE_ON_CLOSE. Otherwise, it is disposed immediately.
	 */
	public void remove() {
		if (SystemTray.isSupported()) {
			SystemTray.getSystemTray().remove(trayIcon);
			if (frame.isVisible())
				frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			else
				frame.dispose();
		}
	}

	/* -- Private -- */

	private final JFrame frame;

	private final TrayIcon trayIcon;

	/** The logger for this class. */
	private static Logger logger = Logger.getLogger(TrayController.class
			.getName());

}
