/*
 * Distriblend: A distributed rendering system for Blender Copyright (C) 2007:
 * Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share;

import java.io.File;

/**
 * Represents a map between a file and a relative path. This class is useful for
 * representing file transfers.
 * 
 * @author Ian Thompson
 * 
 */
public class FileMap {

	/* -- Constructors -- */

	/**
	 * Creates a new FileMap instance between the given source and destination
	 * paths.
	 * <p>
	 * There are three valid situations:
	 * <ul>
	 * <li>Both paths are not <code>null</code> and represent normal files.
	 * <li>Both paths are not <code>null</code> and represent directories.
	 * <li>The source is <code>null</code>, the destination is a not
	 * <code>null</code> and the destination represents a directory.
	 * </ul>
	 * In any other case an {@link IllegalArgumentException} is thrown. The last
	 * option is useful for the creation of remote directories, where a source
	 * directory is not required and may be set to <code>null</code>.
	 * 
	 * @param source
	 *            The path to the source file, directory or <code>null</code>.
	 * @param dest
	 *            The relative path to the destination file or directory.
	 * @throws IllegalArgumentException
	 *             If none of the above cases are satisfied.
	 */
	public FileMap(File source, LocalPath dest) throws IllegalArgumentException {
		if (dest == null)
			throw new IllegalArgumentException("Destination cannot be null.");
		if (source == null && !dest.isDirectory())
			throw new IllegalArgumentException(
					"Source may only be null for directory destinations.");
		this.source = source;
		this.dest = dest;
	}

	/* -- Encapsulation methods -- */

	/**
	 * Returns the destination path for this map.
	 * 
	 * @return The destination path for this map.
	 */
	public LocalPath getDestinationPath() {
		return dest;
	}

	/**
	 * Changes the name of the destination. This method does not call
	 * {@link File#renameTo(File)} or make any attempt to rename an actual file.
	 * Instead a new destination is created. The source file and path is
	 * unaffected.
	 * 
	 * @param newName
	 *            The new name for the destination file or directory.
	 * @throws IllegalArgumentException
	 *             If the new name is not a valid file name as dictated by
	 *             {@link SafeString#isSafe(String)}.
	 * @see SafeString#isSafe(String)
	 */
	public void renameDestination(String newName)
			throws IllegalArgumentException {

		if (!SafeString.isSafe(newName))
			throw new IllegalArgumentException(
					"The newName parameter contains invalid characters.");

		LocalPath parent = dest.getParent();
		if (parent == null)
			dest = new LocalPath(newName, dest.isDirectory());
		else
			dest = new LocalPath(parent, newName, dest.isDirectory());
	}

	/**
	 * Returns the source path for this map.
	 * 
	 * @return The source path for this map.
	 */
	public File getSourcePath() {
		return source;
	}

	/**
	 * Returns <code>true</code> if the destination path represents a
	 * directory.
	 * 
	 * @return <code>true</code> if the destination path represents a
	 *         directory.
	 */
	public boolean isDirectory() {
		return dest.isDirectory();
	}

	/* -- Overridden methods -- */

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		if (isDirectory())
			return dest.getName();
		else
			return dest.getName() + " (" + source.getName() + ")";
	}

	/* -- Private -- */

	/**
	 * Source file path.
	 */
	private final File source;

	/**
	 * Destination relative path.
	 */
	private LocalPath dest;
}
