/*
 * Distriblend: A distributed rendering system for Blender Copyright (C) 2007:
 * Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share;

/**
 * Enumerates the image formats supported by the system and provides their names
 * and file extensions.
 * 
 * @author Ian Thompson
 * 
 */
public enum ImageFormat {

	/** Portable network graphics image format. */
	PNG("Portable Network Graphics", "PNG", new String[] { "png" }),

	/** JPEG image format. */
	JPEG("JPEG Image", "JPEG", new String[] { "jpg", "jpeg" }),

	/** TIFF image format. */
	TIFF("TIFF Image", "TIFF", new String[] { "rgb", "tif", "tiff" }),

	/** Targa image format. */
	TARGA("Targa Image", "TGA", new String[] { "tga" }),

	/** Windows Bitmap image format. */
	BMP("Windows Bitmap", "BMP", new String[] { "bmp" });

	/**
	 * The friendly name of this format.
	 */
	public final String name;

	/**
	 * The name recognised by the renderer.
	 */
	public final String codeName;

	/**
	 * An array of associated file extensions for this format.
	 */
	public final String[] extensions;

	/**
	 * Attempts to find and return a format matching the given extension.
	 * 
	 * @param extension
	 *            The extension to search for.
	 * @return The format with the matching extension.
	 * @throws IllegalArgumentException
	 *             If no match is found <em>and</em> if
	 *             {@link #valueOf(String)} throws an
	 *             <code>IllegalArgumentException</code>.
	 */
	public static ImageFormat fromExtension(String extension)
			throws IllegalArgumentException {
		for (ImageFormat imgfmt : ImageFormat.values())
			for (String ext : imgfmt.extensions)
				if (ext.equalsIgnoreCase(extension))
					return imgfmt;
		return valueOf(extension.toUpperCase());
	}

	/**
	 * Private constructor to initialize the name and extension.
	 * 
	 * @param name
	 *            The format name.
	 * @param codeName
	 *            The name as recognised by the renderer.
	 * @param extensions
	 *            The file extension for this format.
	 */
	private ImageFormat(String name, String codeName, String[] extensions) {
		this.name = name;
		this.codeName = codeName;
		this.extensions = extensions;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Enum#toString()
	 */
	@Override
	public String toString() {
		String ext = "";
		for (String s : extensions)
			ext += "*." + s + ", ";
		return this.name + " (" + ext.substring(0, ext.length() - 2) + ")";
	}
}
