/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share;

import java.util.StringTokenizer;

/**
 * A utility class for validating {@link String}s.
 * 
 * @author Ian Thompson
 * 
 */
public class SafeString {

	/**
	 * Tests whether the specified {@link String} is safe for use as a network
	 * or disk path component on any system.
	 * <p>
	 * Please note that some values that would be acceptable on certain
	 * platforms may not be on others and as such these values will be reported
	 * unsafe.
	 * 
	 * @param s
	 *            The string to test.
	 * @return Whether or not the specified {@link String} is safe for use in
	 *         file or network path components.
	 */
	public static boolean isSafe(String s) {
		if (s == "" || s == "." || s == "..")
			return false;

		char[] chars = s.toCharArray();
		for (char ch : chars) {
			if (ch < 32)
				return false;
			switch (ch) {
			case '/':
			case '\\':
			case ':':
			case '*':
			case '?':
			case '<':
			case '>':
			case '|':
			case '':
			case '%':
			case '"':
			case '\r':
			case '\n':
			case '\t':
				return false;
			}
		}
		return true;
	}

	/**
	 * Calls {@link #isSafe(String)} on each path component.
	 * 
	 * @param path
	 * @return Whether or not the specified string is safe for use as a file or
	 *         network path.
	 */
	public static boolean isPathSafe(String path) {
		String token = "\\";
		if (path.indexOf('/') >= 0)
			token = "/";
		if (path.startsWith(token))
			return false;
		StringTokenizer st = new StringTokenizer(path, token);
		while (st.hasMoreTokens()) {
			if (!isSafe(st.nextToken()))
				return false;
		}
		return true;
	}

	/**
	 * The value of this safe string.
	 */
	public final String value;

	/**
	 * 
	 * @param value
	 */
	public SafeString(String value) {
		if (!isSafe(value))
			throw new IllegalArgumentException("String not safe.");
		this.value = value;
	}
}
