/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

import net.sf.distriblend.share.event.SettingChangedEvent;
import net.sf.distriblend.share.event.SettingChangedListener;

/**
 * Presents a checked system for setting, getting, loading and storing a group
 * of settings. A description may be added for each setting. This makes the file
 * produced by the {@link #save(File)} method easier to edit outside of the
 * application.
 * 
 * @author Ian Thompson
 * 
 */
public class Settings {

	/**
	 * Returns a single instance appropriate to the current application.
	 * 
	 * @return The single {@link Settings} instance.
	 */
	public static Settings instance() {
		if (instance == null) {
			instance = new Settings();
		}
		return instance;
	}

	/**
	 * Returns the description for the specified setting.
	 * 
	 * @param key
	 *            The key of the setting.
	 * @return The description for the setting.
	 */
	public String getDescription(String key) {
		return settingsMap.get(key).description;
	}

	/**
	 * Sets the description for the specified setting. The setting must already
	 * exist for this operation to be performed, otherwise a
	 * {@link NullPointerException} will be thrown.
	 * 
	 * @param key
	 *            The key of the setting.
	 * @param description
	 *            The description for the setting.
	 */
	public void setDescription(String key, String description) {
		Setting setting = settingsMap.get(key);
		if (setting == null)
			throw new IllegalArgumentException("No setting " + key);
		setting.description = description;
	}

	/**
	 * Searches for the key in the internal set of properties. If the key is
	 * found the corresponding value is returned as a {@link String}. Otherwise
	 * an exception is thrown.
	 * 
	 * @param key
	 *            The key to search for.
	 * @return The matching property value converted to a {@link String}.
	 * @throws IllegalArgumentException
	 *             If no property is found with a matching key.
	 */
	public synchronized String getStringProperty(String key)
			throws IllegalArgumentException {
		String value = get(key);
		if (value == null)
			throw new IllegalArgumentException("No setting " + key);
		return value;
	}

	/**
	 * Searches for the key in the internal set of properties. If the key is
	 * found the corresponding value is returned as an <code>int</code>.
	 * Otherwise an exception is thrown.
	 * 
	 * @param key
	 *            The key to search for.
	 * @return Property value parsed as an <code>int</code>.
	 * @throws IllegalArgumentException
	 *             If no property is found with a matching key.
	 * @throws NumberFormatException
	 *             If a matching property <em>is</em> found but the value is
	 *             not a parsable <code>int</code>.
	 * @see Integer#parseInt(String)
	 */
	public synchronized int getIntProperty(String key)
			throws IllegalArgumentException, NumberFormatException {
		String value = get(key);
		if (value == null)
			throw new IllegalArgumentException("No setting " + key);
		return Integer.parseInt(value);
	}

	/**
	 * Searches for the key in the internal set of properties. If the key is
	 * found the corresponding value is returned as a <code>double</code>.
	 * Otherwise an exception is thrown.
	 * 
	 * @param key
	 *            The key to search for.
	 * @return Property value parsed as a <code>double</code>.
	 * @throws IllegalArgumentException
	 *             If no property is found with a matching key.
	 * @throws NumberFormatException
	 *             If a matching property <em>is</em> found but the value is
	 *             not a parsable <code>double</code>.
	 * @see Double#parseDouble(String)
	 */
	public synchronized double getDoubleProperty(String key)
			throws IllegalArgumentException, NumberFormatException {
		String value = get(key);
		if (value == null)
			throw new IllegalArgumentException("No setting " + key);
		return Double.parseDouble(value);
	}

	/**
	 * Searches for the key in the internal set of properties. If the key is
	 * found the corresponding value is returned as a <code>boolean</code>.
	 * Otherwise an exception is thrown.
	 * 
	 * @param key
	 *            The key to search for.
	 * @return Property value parsed as a <code>boolean</code>.
	 * @throws IllegalArgumentException
	 *             If no property is found with a matching key.
	 * @throws NumberFormatException
	 *             If a matching property <em>is</em> found but the value is
	 *             not a parsable <code>boolean</code>.
	 * @see Boolean#parseBoolean(String)
	 */
	public synchronized boolean getBooleanProperty(String key)
			throws IllegalArgumentException, NumberFormatException {
		String value = get(key);
		if (value == null)
			throw new IllegalArgumentException("No setting " + key);
		return Boolean.parseBoolean(value);
	}

	/**
	 * Sets the property value associated with the specified key to the given
	 * value. If the property key does not yet exist it will be created.
	 * 
	 * @param key
	 *            The property to set.
	 * @param value
	 *            The new value.
	 */
	public synchronized void setStringProperty(String key, String value) {
		set(key, value);
	}

	/**
	 * Sets the property value associated with the specified key to the given
	 * value. If the property key does not yet exist it will be created.
	 * 
	 * @param key
	 *            The property to set.
	 * @param value
	 *            The new value.
	 */
	public synchronized void setIntProperty(String key, int value) {
		set(key, "" + value);
	}

	/**
	 * Sets the property value associated with the specified key to the given
	 * value. If the property key does not yet exist it will be created.
	 * 
	 * @param key
	 *            The property to set.
	 * @param value
	 *            The new value.
	 */
	public synchronized void setDoubleProperty(String key, double value) {
		set(key, "" + value);
	}

	/**
	 * Sets the property value associated with the specified key to the given
	 * value. If the property key does not yet exist it will be created.
	 * 
	 * @param key
	 *            The property to set.
	 * @param value
	 *            The new value.
	 */
	public synchronized void setBooleanProperty(String key, boolean value) {
		set(key, "" + value);
	}

	/**
	 * Returns whether the internal property set contains the specified key.
	 * 
	 * @param key
	 * @return Whether the specified property key has been defined.
	 */
	public synchronized boolean hasProperty(String key) {
		return settingsMap.containsKey(key);
	}

	/**
	 * Returns the number of properties defined.
	 * 
	 * @return The number of properties defined.
	 */
	public synchronized int count() {
		return settingsMap.size();
	}

	/**
	 * Returns an {@link Iterator} for iterating over the property keys.
	 * 
	 * @return An {@link Iterator} for iterating over the property keys.
	 */
	public Iterator<String> keys() {
		return settingsMap.keySet().iterator();
	}

	/**
	 * Loads a set of settings pairs from the specified file. The format of this
	 * file must be a line separated list of <code>key=value</code> pairs
	 * where each pair may be optionally preceded by one or more description
	 * lines. A description line must start with a <code>#</code> character.
	 * <p>
	 * Blank lines and lines starting with <code>##</code> are ignored and may
	 * not be written back to the file when saving.
	 * <p>
	 * The following lists an example file: <blockquote>
	 * 
	 * <pre>
	 * ## Distriblend Settings File ##
	 * 
	 * # The address to which the server should bind for incoming connections
	 * server.address=192.168.1.3
	 * 
	 * # The port to listen on for incoming connections
	 * server.port=20000
	 * 
	 * ...
	 * </pre>
	 * 
	 * </blockquote>
	 * 
	 * @param file
	 *            The file from which properties should be loaded.
	 * @throws FileNotFoundException
	 *             If the specified file does not exist, is a directory or
	 *             cannot be opened for reading.
	 * @throws IOException
	 *             If an error occurs reading the properties file.
	 */
	public synchronized void load(File file) throws FileNotFoundException,
			IOException {
		BufferedReader in = new BufferedReader(new FileReader(file));
		String line;
		StringBuffer descBuffer = new StringBuffer();
		int equals;
		Setting setting;
		while ((line = in.readLine()) != null) {
			line = line.trim();
			if (line.equals("") || line.startsWith("##"))
				continue;
			if (line.startsWith("#")) {
				descBuffer.append(line.substring(1).trim());
				descBuffer.append(" ");
			} else {
				equals = line.indexOf('=');
				if (equals < 1)
					throw new IllegalArgumentException(
							"Malformed settings format.");
				setting = new Setting();
				setting.description = descBuffer.toString();
				setting.value = line.substring(equals + 1).trim();
				settingsMap.put(line.substring(0, equals).trim(), setting);
				descBuffer = new StringBuffer();
			}
		}
	}

	/**
	 * Saves the internal settings map to the specified file.
	 * 
	 * @param file
	 *            The files to which the settings should be written.
	 * @throws IOException
	 *             If an error occurs while writing to the file.
	 */
	public synchronized void save(File file) throws IOException {
		PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(
				file)));
		Set<Entry<String, Setting>> entries = settingsMap.entrySet();
		String key;
		Setting setting;
		int start, end;
		String desc;

		out.println("## Distriblend Settings File ##");
		out.println();
		for (Entry<String, Setting> entry : entries) {
			key = entry.getKey();
			setting = entry.getValue();
			desc = setting.description;

			/* Print the description lines, each preceded by # */
			if (desc != null) {

				/* Get a line's worth, move left until word gap, chop there. */
				for (start = 0, end = 79; end <= desc.length(); end += 79) {
					end = start + desc.substring(start, end).lastIndexOf(' ');
					if (end == -1)
						break;
					out.println("# " + desc.substring(start, end));
					start = end + 1;
				}
				out.println("# " + desc.substring(start));
			}
			/* Print the key=value pair */
			out.println(key + "=" + setting.value);
			out.println();
			out.flush();
		}
		out.close();
	}

	/**
	 * Adds the specified listener to the list of listeners set to receive
	 * notification of setting changes.
	 * 
	 * @param l
	 *            The listener to add.
	 */
	public void addSettingChangedListener(SettingChangedListener l) {
		listeners.add(l);
	}

	/**
	 * Removes the specified listener from the list of listeners set to receive
	 * notification of setting changes.
	 * 
	 * @param l
	 *            The listener to remove.
	 */
	public void removeSettingChangedListener(SettingChangedListener l) {
		listeners.remove(l);
	}

	/**
	 * Fires an event to indicate that the specified setting key has changed.
	 * The old and new values are used to create the event.
	 * 
	 * @param key
	 *            The key of the setting that has changed.
	 * @param oldValue
	 *            The old value.
	 * @param newValue
	 *            The new value.
	 */
	protected void fireSettingChanged(String key, String oldValue,
			String newValue) {
		SettingChangedEvent event = new SettingChangedEvent(key, oldValue,
				newValue);
		for (SettingChangedListener l : listeners)
			l.settingChanged(event);
	}

	/**
	 * The single instance of the {@link Settings} class.
	 */
	private static Settings instance = null;

	/**
	 * Internal {@link Properties} set.
	 */
	private final TreeMap<String, Setting> settingsMap;

	/**
	 * The set of listeners.
	 */
	private final Vector<SettingChangedListener> listeners;

	/**
	 * Creates the single instance.
	 */
	private Settings() {
		this.settingsMap = new TreeMap<String, Setting>();
		this.listeners = new Vector<SettingChangedListener>();
	}

	/**
	 * Internal command for setting properties and if necessary automatically
	 * saving the changes. This method also fires the necessary events.
	 * 
	 * @param key
	 *            The key of the property to set.
	 * @param newValue
	 *            The new value to assign.
	 */
	private void set(String key, String newValue) {
		Setting setting = settingsMap.get(key);
		if (setting == null) {
			setting = new Setting();
			settingsMap.put(key, setting);
		}
		String oldValue = setting.value;
		setting.value = newValue;
		fireSettingChanged(key, oldValue, newValue);
	}

	/**
	 * Internal command for the retrieval of settings. Simply wraps the
	 * {@link Properties#getProperty(String)} method.
	 * 
	 * @param key
	 *            The key of the property to retrieve.
	 * @return The property value or <code>null</code> if no matching property
	 *         was found.
	 */
	private String get(String key) {
		return settingsMap.get(key).value;
	}

	/**
	 * Container type for settings and their descriptions.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class Setting {
		String description = null;
		String value = null;
	}
}
