/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share;

import java.io.Serializable;
import java.util.Queue;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * A process to be carried out by one of a set of nodes. It holds all the
 * information required, given the correct set of input files, to render and
 * produce a set of rendered frames.
 * <p>
 * Typically, a task is created, amended by adding frames, sent to a distributor
 * (where further details are added) and then sent to an available node. Once it
 * reaches the node it is instructed to render a single frame. Nodes may then be
 * instructed to render further frames from the same task or a new task.
 * 
 * @author Ian Thompson
 * 
 */
public class Task implements Serializable {

	/* -- Constructors -- */

	/**
	 * Creates a new <code>Task</code> instance with the given source file and
	 * image format. Frames should be added to the task using the
	 * {@link #addFrame(int)} and {@link #addRange(int, int)} methods.
	 * 
	 * @param sourceFile
	 * @param imageFormat
	 */
	public Task(LocalPath sourceFile, ImageFormat imageFormat) {
		this.sourceFile = sourceFile;
		this.imageFormat = imageFormat;
	}

	/**
	 * Creates a new <code>Task</code> instance with the default image format.
	 * A task produced by this constructor initially has no frames to render.
	 * 
	 * @param sourceFile
	 * 
	 * @see #addFrame(int)
	 * @see #addRange(int, int)
	 */
	public Task(LocalPath sourceFile) {
		this(sourceFile, ImageFormat.PNG);
	}

	/**
	 * Creates a new <code>Task</code> instance with the default image format
	 * and given set of frames to render.
	 * 
	 * @param sourceFile
	 * @param startFrame
	 * @param endFrame
	 */
	public Task(LocalPath sourceFile, int startFrame, int endFrame) {
		this(sourceFile);
		this.addRange(startFrame, endFrame);
	}

	/**
	 * Creates a new <code>Task</code> instance with the given set of frames
	 * to render in the image format specified.
	 * 
	 * @param sourceFile
	 * @param imageFormat
	 * @param startFrame
	 * @param endFrame
	 */
	public Task(LocalPath sourceFile, ImageFormat imageFormat, int startFrame,
			int endFrame) {
		this(sourceFile, imageFormat);
		this.addRange(startFrame, endFrame);
	}

	/* -- Encapsulation methods -- */

	/**
	 * Returns the relative path of the source file for this task.
	 * 
	 * @return the relative path of the source file for this task.
	 */
	public LocalPath getSourceFile() {
		return sourceFile;
	}

	/**
	 * Returns the format this task is set to use for rendering frames.
	 * 
	 * @return the format this task is set to use for rendering frames.
	 */
	public ImageFormat getImageFormat() {
		return imageFormat;
	}

	/**
	 * Adds a range to the set of frame to render. If <code>startFrame</code>
	 * is strictly less than <code>endFrame</code>, this method does nothing.
	 * 
	 * @param startFrame
	 *            The first index of the range to add.
	 * @param endFrame
	 *            The final index of the range to add.
	 */
	public void addRange(int startFrame, int endFrame) {
		addRange(startFrame, endFrame, 1);
	}

	/**
	 * Adds a range to the set of frame to render. If <code>startFrame</code>
	 * is strictly less than <code>endFrame</code>, this method does nothing.
	 * 
	 * @param startFrame
	 *            The first index of the range to add.
	 * @param endFrame
	 *            The final index of the range to add.
	 * @param step
	 *            The incrementing step when adding frames. Setting
	 *            <code>step</code> greater than <code>1</code> will skip
	 *            <code>step&nbsp;-&nbsp;1</code> frames.
	 */
	public void addRange(int startFrame, int endFrame, int step) {
		validateFrameIndex(startFrame);
		validateFrameIndex(endFrame);
		if (step < 1)
			throw new IllegalArgumentException("Step must be positive.");
		for (int i = startFrame; i <= endFrame; i += step)
			this.frames.add(new Integer(i));
	}

	/**
	 * Adds a frame to the set of frames to render.
	 * 
	 * @param frame
	 *            The index of the frame to add.
	 */
	public void addFrame(int frame) {
		validateFrameIndex(frame);
		this.frames.add(new Integer(frame));
	}

	/**
	 * Attempts to check out a frame for rendering.
	 * 
	 * @return The frame number to render or -1 if there are no frames left
	 *         checked in.
	 */
	public synchronized int checkOutFrame() {
		/* Set up check out system */
		if (unprocessedFrames == null || checkedOutFrames == null) {
			unprocessedFrames = new ConcurrentLinkedQueue<Integer>();
			checkedOutFrames = new TreeSet<Integer>();
			for (Integer f : frames)
				if (!completedFrames.contains(f))
					unprocessedFrames.add(f);
		}

		/* Check a frame out */
		Integer i = unprocessedFrames.poll();
		if (i == null)
			return -1;
		checkedOutFrames.add(i);
		return i.intValue();
	}

	/**
	 * Checks a frame back in after processing with a status flag reflecting
	 * whether the frame rendered successfully or failed.
	 * 
	 * @param frame
	 *            The frame number.
	 * @param success
	 *            The success status of the rendering.
	 */
	public synchronized void checkIn(int frame, boolean success) {
		if (!checkedOutFrames.remove(frame))
			throw new IndexOutOfBoundsException("Frame not checked out for "
					+ "rendering.");

		this.completedFrames.add(frame);
		// TODO: Support okay/failed frames (success parameter).
	}

	/**
	 * Checks a frame back in as if it were never checked out.
	 * 
	 * @param frame
	 *            The frame number.
	 */
	public synchronized void cancel(int frame) {
		if (!checkedOutFrames.remove(frame))
			throw new IndexOutOfBoundsException("Frame not checked out for "
					+ "rendering.");

		this.unprocessedFrames.add(frame);
	}

	/**
	 * Returns the number of number of frames involved in this task.
	 * 
	 * @return The number of frames this task is to produce.
	 */
	public int getFrameCount() {
		return frames.size();
	}

	/**
	 * Returns an array of this task's set of frames.
	 * 
	 * @return An array of this task's set of frames.
	 */
	public int[] getFrames() {
		synchronized (frames) {
			int[] frames = new int[this.frames.size()];
			int i = 0;
			for (Integer frame : this.frames) {
				frames[i++] = frame.intValue();
			}
			return frames;
		}
	}

	/**
	 * Returns the number of number of frames rendered.
	 * 
	 * @return The number of frames rendered.
	 */
	public int getCompletedFrameCount() {
		return completedFrames.size();
	}

	/**
	 * Returns an array of this task's set of completed frames.
	 * 
	 * @return An array of this task's set of completed frames.
	 */
	public int[] getCompletedFrames() {
		synchronized (completedFrames) {
			int[] frames = new int[this.completedFrames.size()];
			int i = 0;
			for (Integer frame : this.completedFrames) {
				frames[i++] = frame.intValue();
			}
			return frames;
		}
	}

	/**
	 * Returns <code>true</code> if all frames have been processed.
	 * 
	 * @return Whether all frames have been processed.
	 */
	public boolean isComplete() {
		return completedFrames.size() == frames.size();
	}

	/**
	 * Returns the name of the user to which this task belongs.
	 * 
	 * @return The name of the user to which this task belongs.
	 */
	public String getUser() {
		return user;
	}

	/**
	 * Sets the user to which this task belongs.
	 * 
	 * @param user
	 *            The name of the user to which this task belongs.
	 */
	public void setUser(String user) {
		this.user = user;
	}

	/**
	 * Returns the name of the project to which this task belongs.
	 * 
	 * @return The name of the project to which this task belongs.
	 */
	public String getProjectName() {
		return projectName;
	}

	/**
	 * Sets the project to which this task belongs.
	 * 
	 * @param projectName
	 *            The name of the project to which this task belongs.
	 */
	public void setProjectName(String projectName) {
		this.projectName = projectName;
	}

	/* -- Overridden methods -- */

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Task: " + sourceFile.getName() + ", frames: " + ranges();
	}

	/* -- Private -- */

	/**
	 * A set of frames to render. Once on the distributor this collection is
	 * copied to the unprocessedFrames collection. The framesArray field is used
	 * to serialize this transient field.
	 */
	private final TreeSet<Integer> frames = new TreeSet<Integer>();

	/** Frame numbers are copied here when the task reaches the distributor. */
	private transient Queue<Integer> unprocessedFrames = null;

	/** Frame numbers are moved here when checked out. */
	private transient TreeSet<Integer> checkedOutFrames = null;

	/** Frame numbers are moved here when marked complete. */
	private final TreeSet<Integer> completedFrames = new TreeSet<Integer>();

	/** The relative path to the source file used to initiate rendering. */
	private final LocalPath sourceFile;

	/** The format to use when saving each rendered frame. */
	private final ImageFormat imageFormat;

	/** The name of the user to whom this task belongs. */
	private String user = null;

	/** The name of the project to which this task belongs. */
	private String projectName = null;

	/* -- Serialization -- */

	/**
	 * Converts the {@link TreeSet} of frames to an ordered array of integers
	 * for faster serialization.
	 * 
	 * @param s
	 *            The stream to which this object should be serialized.
	 * @throws java.io.IOException
	 */
	private void writeObject(java.io.ObjectOutputStream s)
			throws java.io.IOException {
		s.defaultWriteObject();
	}

	/**
	 * Fills the {@link TreeSet} of frames with the entries of the stored array
	 * of integers.
	 * 
	 * @param s
	 *            The stream from which this object should be read.
	 * @throws java.io.IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(java.io.ObjectInputStream s)
			throws java.io.IOException, ClassNotFoundException {
		s.defaultReadObject();
	}

	/**
	 * Checks the validity of a frame index.
	 * 
	 * @param index
	 *            The index to check.
	 */
	private void validateFrameIndex(int index) {
		if (index < 1)
			throw new IllegalArgumentException(
					"Frame index must be greater than zero.");
	}

	private String ranges() {
		int i;
		String s = "";
		Integer first = frames.first();
		while (true) {
			for (i = 0; frames.contains(first + i + 1); i++)
				;
			if (i == 0)
				s += first;
			else
				s += first + " to " + (first + i);
			first = frames.higher(first + i);
			if (first != null)
				s += ", ";
			else
				break;
		}
		return s;
	}
}
