/*
 * Distriblend: A distributed rendering system for Blender Copyright (C) 2007:
 * Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share.net;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Vector;

import net.sf.distriblend.share.net.event.ConnectionEvent;
import net.sf.distriblend.share.net.event.ConnectionListener;

/**
 * A connection initiator designed to connect to a {@link Receptor}.
 * 
 * @author Ian Thompson
 * 
 */
public class Connector {

	/* -- Constructor -- */

	/**
	 * 
	 * @param host
	 * @param port
	 * @param timeout
	 */
	public Connector(String host, int port, int timeout) {
		this.host = host;
		this.port = port;
		this.timeout = timeout;
		this.listeners = new Vector<ConnectionListener>();
		this.connected = false;
	}

	/* -- Encapsulation methods -- */

	/**
	 * 
	 * @return Host
	 */
	public String getHost() {
		return host;
	}

	/**
	 * 
	 * @return Port
	 */
	public int getPort() {
		return port;
	}

	/* -- Event system -- */

	/**
	 * 
	 * @param listener
	 */
	public void addConnectionListener(ConnectionListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * 
	 * @param listener
	 */
	public void removeConnectionListener(ConnectionListener listener) {
		this.listeners.remove(listener);
	}

	/**
	 * 
	 * 
	 */
	public void fireConnectionEstablished() {
		for (ConnectionListener l : listeners)
			l.connectionEstablished(new ConnectionEvent(Connector.this));
	}

	/**
	 * 
	 * 
	 */
	public void fireConnectionLost() {
		for (ConnectionListener l : listeners)
			l.connectionLost(new ConnectionEvent(Connector.this));
	}

	/* -- Connection handling methods -- */

	/**
	 * 
	 */
	public void connect() {
		if (socket != null)
			throw new IllegalStateException(
					"Only a single connection is permitted per Connector.");

		socket = new Socket();
		this.watcher = new Watcher();
		this.watcher.start();
	}

	/**
	 * 
	 * 
	 */
	public void close() {
		if (socket != null) {
			try {
				socket.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			watcher.close();
		}
	}

	/**
	 * 
	 * @return <code>true</code> if and only if the connection has been
	 *         established and accepted.
	 */
	public boolean isSuccessfullyConnected() {
		return socket != null && socket.isConnected() && !socket.isClosed()
				&& connected;
	}

	/**
	 * 
	 * @return <code>true</code> if no attempt has been made to connect or an
	 *         attempt failed. If trying to connect, this method will return
	 *         <code>false</code>.
	 */
	public boolean isDisconnected() {
		return socket == null;
	}

	/* -- Data transfer methods -- */

	/**
	 * @param object
	 * @throws IOException
	 */
	public void send(Serializable object) throws IOException {
		try {
			out.writeUnshared(object);
			out.flush();
		} catch (IOException e) {
			close();
			throw e;
		}
	}

	/**
	 * 
	 * @return Object
	 * @throws IOException
	 */
	public Object receive() throws IOException {
		try {
			return in.readObject();
		} catch (ClassNotFoundException e) {
			close();
			throw new IOException(e);
		} catch (IOException e) {
			close();
			throw e;
		}
	}

	/**
	 * @return the in
	 */
	public ObjectInputStream getInputStream() {
		return in;
	}

	/**
	 * @return the out
	 */
	public ObjectOutputStream getOutputStream() {
		return out;
	}

	/* -- Private -- */

	/**
	 * 
	 */
	private final String host;

	/**
	 * 
	 */
	private final int port;

	/**
	 * 
	 */
	private final int timeout;

	/**
	 * 
	 */
	private volatile boolean connected;

	/**
	 * 
	 */
	private Socket socket = null;

	/**
	 * 
	 */
	private ObjectInputStream in = null;

	/**
	 * 
	 */
	private ObjectOutputStream out = null;

	/**
	 * 
	 */
	private final Vector<ConnectionListener> listeners;

	private Watcher watcher;

	/**
	 * Attempts to connect the socket to the configured host and port. If the
	 * connection is established, fails to connect or at a later stage, is
	 * broken, an event is fired to all <code>ConnectionListener</code>s.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class Watcher extends Thread {

		/**
		 * Creates a new <code>Watcher</code> instance for the parent
		 * <code>Connector</code>.
		 */
		public Watcher() {
			super("Connector.Watcher [" + Connector.this + "]");
			this.setDaemon(true);
		}

		/**
		 * Closes the connection.
		 */
		public void close() {
			synchronized (this) {
				closed = true;
				this.notify();
			}
		}

		/**
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			try {
				socket.connect(new InetSocketAddress(host, port), timeout);
				out = new ObjectOutputStream(new BufferedOutputStream(socket
						.getOutputStream()));
				out.flush();
				in = new ObjectInputStream(new BufferedInputStream(socket
						.getInputStream()));
				socket.setSoTimeout(timeout);
				connected = socket.isConnected();
			} catch (IOException e) {
				connected = false;
			}
			if (connected) {
				fireConnectionEstablished();
			}

			while (socket.isConnected() && !closed) {
				try {
					synchronized (this) {
						this.wait();
					}
				} catch (InterruptedException e) {
				}
			}
			connected = false;
			fireConnectionLost();
			socket = null;
		}

		private boolean closed = false;
	}
}
