/*
 * Distriblend: A distributed rendering system for Blender
 * Copyright (C) 2007: Ian Philip Thompson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share.net;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;

/**
 * A simple, generic server for accepting connections and passing them on to the
 * {@link ConnectionHandler} type specified on construction. Single or
 * multi-threaded servers of various types can thus be created using this class.
 * 
 * @author Ian Thompson
 * 
 */
public class Server {

	/**
	 * 
	 * @param port
	 * @param handler
	 */
	public Server(int port, Class<? extends ConnectionHandler> handler) {
		this(null, port, handler);
	}

	/**
	 * 
	 * @param bindAddress
	 * @param port
	 * @param handler
	 */
	public Server(InetAddress bindAddress, int port,
			Class<? extends ConnectionHandler> handler) {
		this.bindAddress = bindAddress;
		this.port = port;
		this.worker = new Worker();
		this.worker.start();
		this.handler = handler;
	}

	/**
	 * Attempts to shutdown this server. If the server has already been shutdown
	 * or has stopped for some other reason, an {@link IllegalStateException} is
	 * thrown.
	 * 
	 * @throws IllegalStateException
	 */
	public void shutdown() throws IllegalStateException {
		this.worker.shutdown();
	}

	/* -- Private -- */

	/**
	 * 
	 */
	private InetAddress bindAddress;

	/**
	 * 
	 */
	private int port;

	/**
	 * Server worker thread.
	 */
	private Worker worker;

	/**
	 * 
	 */
	private Class<? extends ConnectionHandler> handler;

	/**
	 * 
	 * @author Ian Thompson
	 * 
	 */
	private class Worker extends Thread {

		private volatile boolean running;

		private ServerSocket serverSocket = null;

		/**
		 * 
		 */
		public Worker() {
			super("ClientServer$Worker");
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#start()
		 */
		@Override
		public void start() {
			this.running = true;
			super.start();
		}

		/**
		 * 
		 * @return isRunning
		 */
		public boolean isRunning() {
			return running;
		}

		/**
		 * 
		 * @throws IllegalStateException
		 */
		public void shutdown() throws IllegalStateException {
			if (!this.running || this.serverSocket == null)
				throw new IllegalStateException("Server not running.");
			this.running = false;
			try {
				this.serverSocket.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			Socket socket = null;
			try {
				serverSocket = new ServerSocket(port, 50, bindAddress);
				serverSocket.setSoTimeout(3000);
			} catch (IOException e) {
				this.running = false;
				return;
			}
			assert serverSocket != null;
			while (this.running) {
				try {
					socket = serverSocket.accept();
				} catch (SocketTimeoutException e) {
					continue;
				} catch (IOException e) {
					this.running = false;
					break;
				}

				try {
					ConnectionHandler ch = handler.newInstance();
					ch.handleConnection(socket);
				} catch (InstantiationException e) {
					e.printStackTrace();
					this.running = false;
					break;
				} catch (IllegalAccessException e) {
					e.printStackTrace();
					this.running = false;
					break;
				} catch (IOException e) {
					e.printStackTrace();
					break;
				}
			}

			if (socket != null) {
				try {
					socket.close();
				} catch (IOException e) {
				}
			}
		}
	}
}
