/*
 * Distriblend: A distributed rendering system for Blender
 * Copyright (C) 2007: Ian Philip Thompson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share.protocol;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;
import java.io.Serializable;

import net.sf.distriblend.share.SafeString;

/**
 * This class is a container for Authorization protocol messages. It also
 * provides static utility methods for use with this protocol. Authorization
 * takes place in the following manner:
 * <ul>
 * <li>Request: Sent to a server with a user name to initiate authorization.
 * <li>Ready: Reply sent back to the client with handshaking information.
 * <li>Data: Holds authorization data (eg. password) for sending to the server.
 * <li>Response: Final reply indicating whether or not the user was authorized.
 * </ul>
 * 
 * @author Ian Thompson
 * 
 */
public final class Authorization {

	/**
	 * Authorization request command. Used to indicate a request for
	 * authorization of a particular user.
	 * <p>
	 * See {@link Authorization} for detailed usage of this message.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class Request implements Serializable {

		/**
		 * Forms a request for authorization of the specified user.
		 * 
		 * @param user
		 *            The name of the user requesting authroization.
		 * @see Authorization
		 */
		public Request(String user) {
			this.user = user;
			if (!isValid())
				throw new IllegalArgumentException("User name '" + user
						+ "' is invalid.");
		}

		/**
		 * Returns the name of the user requesting authroization.
		 * 
		 * @return The name of the user requesting authroization.
		 */
		public String getUser() {
			return user;
		}

		/* -- Private -- */

		/**
		 * The name of the user requesting authroization.
		 */
		private String user;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			return user != null && !user.equals("") && SafeString.isSafe(user);
		}
	}

	/**
	 * The standard reply to an authorization {@link Request}. This class is a
	 * container for the salt used for hashing an incoming password. In the
	 * future, it may be used to negotiate a more complex authentication or
	 * encryption algorithm.
	 * <p>
	 * See {@link Authorization} for detailed usage of this message.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class Ready implements Serializable {

		/**
		 * Creates a reply to contain the specified salt String.
		 * 
		 * @param salt
		 *            The salt String to use for hashing an incoming password.
		 */
		public Ready(String salt) {
			this.salt = salt;
		}

		/**
		 * Returns the salt that should be used when hashing the user's
		 * password.
		 * 
		 * @return The salt to use for password hashing.
		 */
		public String getSalt() {
			return salt;
		}

		/* -- Private -- */

		/**
		 * Password hashing salt String.
		 */
		private String salt;
	}

	/**
	 * A container for a hashed password.
	 * <p>
	 * See {@link Authorization} for detailed usage of this message.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class Data implements Serializable {

		/**
		 * Creates a container for hashed password data.
		 * 
		 * @param hashedPass
		 *            The hashed password data.
		 */
		public Data(String hashedPass) {
			this.hashedPass = hashedPass;
		}

		/**
		 * Returns the hashed password contained by this message.
		 * 
		 * @return The hashed password.
		 */
		public String getHashedPass() {
			return hashedPass;
		}

		/* -- Private -- */

		/**
		 * A hashed password.
		 */
		private String hashedPass;
	}

	/**
	 * The final response sent to indicate whether the user, specified earlier
	 * in a {@link Request} message, was successfully authorized or not.
	 * <p>
	 * See {@link Authorization} for detailed usage of this message.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class Response implements Serializable {

		/**
		 * Creates a message to send in response to an authorization
		 * {@link Request}. This is the final message of the protocol used to
		 * indicate whether or not a user was authorized.
		 * 
		 * @param accepted
		 */
		public Response(boolean accepted) {
			this.accepted = accepted;
		}

		/**
		 * Returns the status of the authorization.
		 * 
		 * @return <code>true</code> if the user has been authorized,
		 *         <code>false</code> otherwise.
		 */
		public boolean isAccepted() {
			return accepted;
		}

		/* -- Private -- */

		/**
		 * Whether the user specified earlier by {@link Request} has been
		 * authorized.
		 */
		private boolean accepted;
	}

	/* -- Static methods -- */

	/**
	 * Generates a random string of the specified length.
	 * 
	 * @param length
	 *            The length of string to generate.
	 * @return Salt.
	 */
	public static String generateSalt(int length) {
		byte[] bytes = new byte[length];
		for (int i = 0; i < length; i++)
			bytes[i] = (byte) (Math.random() * 256);
		return new String(bytes);
	}

	/* -- Private -- */

	/** Private constructor to prevent instantiation. */
	private Authorization() {
	}
}
