/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share.protocol;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;

import net.sf.distriblend.share.TaskFrame;
import net.sf.distriblend.share.net.Message;

/**
 * Holds messages for requesting tasks and files from a distributor and for
 * sending back updates and rendered frames.
 * 
 * @author Ian Thompson
 * 
 */
public final class Distribution {

	/**
	 * The processing state of a single frame of a task.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static enum TaskFrameState {
		/** Indicates that the frame is being processed. */
		STARTED,
		/** Indicates that processing of the frame failed. */
		FAILED,
		/** Indicates that the frame has been rendered successfully. */
		COMPLETE;
	}

	/**
	 * A message for requesting a task from the distributor.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class TaskRequest implements Message {

		/**
		 * Creates a message for requesting a task from the distributor.
		 */
		public TaskRequest() {
		}
	}

	/**
	 * Presents a response to a {@link TaskRequest} holding the requested task
	 * or <code>null</code> if none is available. This message also holds the
	 * index of the last uploaded archive. If the node does not possess these
	 * files it should request them prior to the execution of the task.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class TaskResponse implements Message {

		/**
		 * Creates a message for transporting a task to a node and instructing
		 * the node to render a single frame. If no task is available, task
		 * should be set to <code>null</code>.
		 * 
		 * @param taskFrame
		 *            The task and frame to render or <code>null</code> if
		 *            none is available.
		 * @param archiveCount
		 *            The non-negative number of archives uploaded for the
		 *            parent project.
		 */
		public TaskResponse(TaskFrame taskFrame, int archiveCount) {
			this.taskFrame = taskFrame;
			this.archiveCount = archiveCount;
			if (!isValid())
				throw new IllegalArgumentException(
						"Frame and archive indices cannot be negative.");
		}

		/**
		 * Returns the task frame to process or <code>null</code> if none is
		 * available.
		 * 
		 * @return The task to process or <code>null</code> if none is
		 *         available.
		 */
		public TaskFrame getTaskFrame() {
			return taskFrame;
		}

		/**
		 * Returns the number of archives uploaded for the parent project.
		 * 
		 * @return The number of archives uploaded for the parent project.
		 */
		public int getArchiveCount() {
			return archiveCount;
		}

		/** The task and frame to render on this task. */
		private final TaskFrame taskFrame;

		/** The number of archives uploaded for the parent project. */
		private final int archiveCount;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException {
			s.defaultReadObject();
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (taskFrame != null && taskFrame.getFrame() < 0)
				return false;
			if (archiveCount < 0)
				return false;
			return true;
		}
	}

	/**
	 * A message for requesting the archived files required by the task being
	 * processed.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class FilesRequest implements Message {

		/**
		 * Creates a message for requesting the archived files required by the
		 * task being processed.
		 * 
		 * @param archivesPossessed
		 *            The index of the latest archive held by the node.
		 */
		public FilesRequest(int archivesPossessed) {
			this.archivesPossessed = archivesPossessed;
		}

		/**
		 * @return The index of the latest archive held by the node.
		 */
		public int getArchivesPossessed() {
			return archivesPossessed;
		}

		private final int archivesPossessed;
	}

	/**
	 * A response message indicating the number of files to expect during the
	 * transfer.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class FilesResponse implements Message {

		/**
		 * Creates a response message indicating the number of files to expect
		 * during the transfer.
		 * 
		 * @param fileCount
		 *            The number of files to expect during the transfer.
		 */
		public FilesResponse(int fileCount) {
			this.fileCount = fileCount;
		}

		/**
		 * @return The number of files to expect during the transfer.
		 */
		public int getFileCount() {
			return fileCount;
		}

		private final int fileCount;
	}

	// See FileTransfer for file transfer messages

	/**
	 * A message used to update the distributor with the progress of the task.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class TaskUpdate implements Message {

		/**
		 * Creates a message for updating the distributor with the task
		 * progress.
		 * 
		 * @param state
		 *            The index of the latest archive held by the node.
		 */
		public TaskUpdate(TaskFrameState state) {
			this.state = state;
		}

		/**
		 * @return The state of the frame being processed.
		 */
		public TaskFrameState getState() {
			return state;
		}

		private final TaskFrameState state;
	}

	/**
	 * A message to indicate that the update was received successfully.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class UpdateReceived implements Message {

		/**
		 * Creates a message to indicate that the update was received
		 * successfully.
		 */
		public UpdateReceived() {
		}
	}

	/**
	 * A message used to upload a rendered frame at the end of task processing.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class FrameUpload implements Message {

		/**
		 * Creates a message for uploading a rendered frame at the end of a
		 * task.
		 */
		public FrameUpload() {
		}
	}

	// See FileTransfer for file transfer messages

	/**
	 * A message to indicate that the upload was received successfully.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class UploadReceived implements Message {

		/**
		 * Creates a message to indicate that the upload was received
		 * successfully.
		 */
		public UploadReceived() {
		}
	}

	/** Private constructor. */
	private Distribution() {
	}
}
