/*
 * Distriblend: A distributed rendering system for Blender
 * 
 * Copyright (C) 2007: Ian Philip Thompson
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package net.sf.distriblend.share.protocol;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;

import net.sf.distriblend.share.SafeString;
import net.sf.distriblend.share.Task;
import net.sf.distriblend.share.net.Message;

/**
 * Holds messages for creating projects; submitting files and tasks for
 * processing, and receiving updates and frames from those tasks.
 * 
 * @author Ian Thompson
 * 
 */
public final class Submission {

	/**
	 * Used to request a list of projects owned by the current user from the
	 * distributor. An instance of this class is typically created by the
	 * client.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class ListProjectsRequest implements Message {

		/**
		 * Creates a request for the project list of the current user.
		 */
		public ListProjectsRequest() {
		}
	}

	/**
	 * Presents a response to a {@link ListProjectsRequest} containing the list
	 * of projects as requested.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class ListProjectsResponse implements Message {

		/**
		 * Creates a response containing the list of projects owned by the
		 * current user.
		 * 
		 * @param projectNames
		 *            The list of projects for the current user.
		 */
		public ListProjectsResponse(String[] projectNames) {
			this.projectNames = projectNames;
		}

		/**
		 * Returns the list of projects for the current user.
		 * 
		 * @return The list of projects for the current user.
		 */
		public String[] getProjects() {
			return projectNames;
		}

		/* -- Private -- */

		/**
		 * The list of projects for the current user.
		 */
		private final String[] projectNames;
	}

	/**
	 * Used to instruct a server to create a new project.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class CreateProjectRequest implements Message {

		/**
		 * Creates a request for the new project specified to be created for the
		 * current authorized user.
		 * 
		 * @param projectName
		 *            The name of the project to create.
		 */
		public CreateProjectRequest(String projectName) {
			this.projectName = projectName;
			if (!isValid())
				throw new IllegalArgumentException(
						"Project names must not contain any of the characters: / \\ : * ? \" < > |");
		}

		/**
		 * Returns the name of the project to create.
		 * 
		 * @return The name of the project to create.
		 */
		public String getProjectName() {
			return projectName;
		}

		/**
		 * The name of the project to create.
		 */
		private final String projectName;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 * @throws NotSerializableException
		 *             If this instance is found to be invalid due to tampering.
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (!SafeString.isSafe(projectName))
				return false;
			return true;
		}
	}

	/**
	 * Presents a response to a {@link CreateProjectRequest}. This response
	 * contains information about whether a project by the same name already
	 * exists.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class CreateProjectResponse implements Message {

		/**
		 * Creates a response to inform the client whether the project was
		 * successfully created.
		 * 
		 * @param successful
		 *            Indicates that a new project was created. It is usually
		 *            good to make this <code>false</code> whenever
		 *            <code>alreadyExists</code> is <code>true</code>.
		 */
		public CreateProjectResponse(boolean successful) {
			this.successful = successful;
		}

		/**
		 * Returns <code>true</code> if and only if a new project was created.
		 * 
		 * @return Whether the project was created successfully.
		 */
		public boolean isSuccessful() {
			return successful;
		}

		/* -- Private -- */

		/**
		 * Whether the project was created successfully.
		 */
		private final boolean successful;
	}

	/**
	 * Indicates a desire to add files to a particular project. An instance of
	 * this class is typically created by the client, serialized and sent to the
	 * server.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class SubmitFilesRequest implements Message {

		/**
		 * Creates a request to add files to the specified project.
		 * 
		 * @param fileCount
		 *            The number of files to add.
		 */
		public SubmitFilesRequest(int fileCount) {
			this.fileCount = fileCount;
			if (!isValid())
				throw new IllegalArgumentException(
						"File count must not be negative.");
		}

		/**
		 * Returns the number of files to add to the project.
		 * 
		 * @return The number of files to add to the project.
		 */
		public int getFileCount() {
			return fileCount;
		}

		/* -- Private -- */

		/**
		 * The number of files to add.
		 */
		private final int fileCount;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 * @throws NotSerializableException
		 *             If this instance is found to be invalid due to tampering.
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException, NotSerializableException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (fileCount < 0)
				return false;
			return true;
		}
	}

	// See FileTransfer for file transfer details

	/**
	 * Presents a response to a {@link SubmitFilesRequest}, indicating the
	 * success of the file upload.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class SubmitFilesResponse implements Message {

		/**
		 * Creates a response for delivering the success status of the file
		 * upload.
		 * 
		 * @param successful
		 *            Whether the upload completed successfully and the files
		 *            were all received.
		 */
		public SubmitFilesResponse(boolean successful) {
			this.successful = successful;
		}

		/**
		 * Returns <code>true</code> if the upload completed successfully and
		 * the files were all received. Otherwise <code>false</code> is
		 * returned.
		 * 
		 * @return Whether the upload completed successfully and the files were
		 *         all received.
		 */
		public boolean getSuccessful() {
			return successful;
		}

		/* -- Private -- */

		/**
		 * Whether the upload completed successfully and the files were all
		 * received.
		 */
		private final boolean successful;
	}

	/**
	 * Indicates a desire to add a set of tasks to a particular project.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class SubmitTasksRequest implements Message {

		/**
		 * Creates a request to add {@link Task}s to the specified project.
		 * 
		 * @param tasks
		 *            A set of tasks to add to the project.
		 */
		public SubmitTasksRequest(Task[] tasks) {
			this.tasks = tasks;
			if (!isValid())
				throw new IllegalArgumentException("Tasks cannot be null.");
		}

		/**
		 * Returns the set of tasks to add to the project.
		 * 
		 * @return The set of tasks to add to the project.
		 */
		public Task[] getTasks() {
			return tasks;
		}

		/* -- Private -- */

		/**
		 * The set of tasks to add.
		 */
		private final Task[] tasks;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 * @throws NotSerializableException
		 *             If this instance is found to be invalid due to tampering.
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException, NotSerializableException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (tasks == null)
				return false;
			return true;
		}
	}

	/**
	 * Presents a response to a {@link SubmitTasksRequest}.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class SubmitTasksResponse implements Message {

		/**
		 * Creates a response for the task upload.
		 */
		public SubmitTasksResponse() {
		}
	}

	/**
	 * Used to request the status of projects owned by the current user. An
	 * instance of this class is typically created by the client, serialized and
	 * sent to the server.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class MonitorTasksRequest implements Message {

		/**
		 * Creates a request for the status of projects owned by the current
		 * user.
		 */
		public MonitorTasksRequest() {
		}
	}

	/**
	 * Presents a response to a {@link MonitorTasksRequest} containing the
	 * number of {@link MonitorItem}s to expect.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class MonitorTasksResponse implements Message {

		/**
		 * Creates a response to hold the specified status items.
		 * 
		 * @param count
		 *            The number of monitor items to expect.
		 */
		public MonitorTasksResponse(int count) {
			this.count = count;
		}

		/**
		 * Returns the number of monitor items to expect.
		 * 
		 * @return The number of monitor items to expect.
		 */
		public int getCount() {
			return count;
		}

		/* -- Private -- */

		/**
		 * The number of projects owned by the current user.
		 */
		private final int count;
	}

	/**
	 * A holder class for communicating the status of a project between systems.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class MonitorItem implements Message,
			Comparable<MonitorItem> {

		/**
		 * The name of the project.
		 */
		public final String projectName;

		/**
		 * The total number of frames that should be rendered for this project.
		 */
		public final int framesToRender;

		/**
		 * The number of frames that have been rendered.
		 */
		public final int framesRendered;

		/**
		 * The paused status of the project.
		 */
		public final boolean paused;

		/**
		 * The amount of time given to processing the project so far. This time
		 * is calculated by each node when running each task and is used to give
		 * an estimated running time.
		 */
		public final int processingTime;

		/**
		 * Creates a new monitoring object to communicate the current status of
		 * the project specified.
		 * 
		 * @param projectName
		 *            The name of the project.
		 * @param framesToRender
		 *            The total number of frames that should be rendered for
		 *            this project.
		 * @param framesRendered
		 *            The number of frames that have been rendered.
		 * @param paused
		 *            The paused status of the project.
		 * @param processingTime
		 *            The amount of time given to processing the project so far.
		 */
		public MonitorItem(String projectName, int framesToRender,
				int framesRendered, boolean paused, int processingTime) {
			this.projectName = projectName;
			this.framesToRender = framesToRender;
			this.framesRendered = framesRendered;
			this.paused = paused;
			this.processingTime = processingTime;
		}

		/**
		 * {@inheritDoc}
		 * 
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(MonitorItem o) {
			return projectName.compareTo(o.projectName);
		}
	}

	/**
	 * Used to instruct a server to supply the client with the frames from a
	 * project.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class FramesRequest implements Message {

		/**
		 * Creates a request for the frames of the specified project.
		 * 
		 * @param projectName
		 *            The name of the project from which the frames should be
		 *            taken (provided it exists).
		 */
		public FramesRequest(String projectName) {
			this.projectName = projectName;
			if (!isValid())
				throw new IllegalArgumentException(
						"Project names must not contain any of the characters: / \\ : * ? \" < > |");
		}

		/**
		 * Returns the name of the project from which frames should be taken.
		 * 
		 * @return The name of the project from which frames should be taken.
		 */
		public String getProjectName() {
			return projectName;
		}

		/* -- Private -- */

		/**
		 * The name of the project from which frames should be taken.
		 */
		private final String projectName;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 * @throws NotSerializableException
		 *             If this instance is found to be invalid due to tampering.
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException, NotSerializableException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (!SafeString.isSafe(projectName))
				return false;
			return true;
		}
	}

	/**
	 * Presents a response to a {@link FramesRequest} indicating the
	 * availability of frames.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class FramesResponse implements Message {

		/**
		 * Creates a response to inform the client of how many frames the server
		 * has from the project (specified in a prior {@link FramesRequest}).
		 * 
		 * @param frameCount
		 *            The number of frames ready for download. Set to
		 *            <code>0</code> to indicate that there are no frames or
		 *            that the project does not exist.
		 */
		public FramesResponse(int frameCount) {
			this.frameCount = frameCount;
			if (!isValid())
				throw new IllegalArgumentException("Frame count must be >= 0.");
		}

		/**
		 * Returns the number of frames ready for download.
		 * 
		 * @return The number of frames ready for download.
		 */
		public int getFrameCount() {
			return frameCount;
		}

		/* -- Private -- */

		/**
		 * The number of frames ready for download.
		 */
		private final int frameCount;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks the fields for this response are valid and suitable. If so,
		 * <code>true</code> is returned. Otherwise, <code>false</code> is
		 * returned.
		 * 
		 * @return Whether all fields are valid and suitable for the response.
		 */
		private boolean isValid() {
			if (frameCount < 0)
				return false;
			return true;
		}
	}

	/**
	 * Used to instruct a server to delete the specified project.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class DeleteRequest implements Message {

		/**
		 * Creates a request to delete the specified project.
		 * 
		 * @param projectName
		 *            The name of the project to delete.
		 */
		public DeleteRequest(String projectName) {
			this.projectName = projectName;
			if (!isValid())
				throw new IllegalArgumentException(
						"Project names must not contain any of the characters: / \\ : * ? \" < > |");
		}

		/**
		 * Returns the name of the project to delete.
		 * 
		 * @return The name of the project to delete.
		 */
		public String getProjectName() {
			return projectName;
		}

		/* -- Private -- */

		/**
		 * The name of the project to delete.
		 */
		private final String projectName;

		/**
		 * Reads and validates an instance of this class from the stream
		 * provided.
		 * 
		 * @param s
		 *            The stream from which to read the object.
		 * @throws IOException
		 *             If an error occurs while reading from the stream.
		 * @throws ClassNotFoundException
		 * @throws NotSerializableException
		 *             If this instance is found to be invalid due to tampering.
		 */
		private void readObject(ObjectInputStream s) throws IOException,
				ClassNotFoundException, NotSerializableException {
			s.defaultReadObject();

			/* Check for tampering since constructor validation. */
			if (!isValid())
				throw new NotSerializableException(this.getClass().getName());
		}

		/**
		 * Checks that the fields for this request are valid and suitable. If
		 * so, <code>true</code> is returned. Otherwise, <code>false</code>
		 * is returned.
		 * 
		 * @return Whether all fields are valid and suitable for this request.
		 */
		private boolean isValid() {
			if (!SafeString.isSafe(projectName))
				return false;
			return true;
		}
	}

	/**
	 * Presents a response to a {@link DeleteRequest}.
	 * 
	 * @author Ian Thompson
	 * 
	 */
	public static final class DeleteResponse implements Message {

		/**
		 * Creates a response to a {@link DeleteRequest}.
		 * 
		 * @param success
		 *            Whether the delete succeeded.
		 */
		public DeleteResponse(boolean success) {
			this.success = success;
		}

		/**
		 * @return Whether the delete succeeded.
		 */
		public boolean getSuccess() {
			return success;
		}

		private final boolean success;
	}

	/** Private constructor. */
	private Submission() {
	}
}
